# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2016, Thibault Saunier <tsaunier@gnome.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the utils.pipeline module."""
# pylint: disable=protected-access,no-self-use
from unittest import mock

from gi.repository import GES
from gi.repository import GLib
from gi.repository import Gst

from pitivi.utils.pipeline import MAX_RECOVERIES
from pitivi.utils.pipeline import Pipeline
from pitivi.utils.pipeline import SimplePipeline
from tests import common


class TestPipeline(common.TestCase):
    """Tests for the Pipeline class."""

    @staticmethod
    def post_fake_error_message(pipeline):
        """Simulates an error message on the specified Pipeline."""
        message = mock.Mock()
        message.type = Gst.MessageType.ERROR
        error = GLib.Error.new_literal(Gst.core_error_quark(),
                                       "fake", Gst.CoreError.TOO_LAZY)
        message.parse_error = mock.MagicMock(return_value=(error, "fake"))
        pipeline._bus_message_cb(None, message)

    def test_recovery(self):
        """Checks the recovery mechanism."""
        pipe = Pipeline(common.create_pitivi_mock())
        pipe.set_timeline(GES.Timeline())

        pipeline_died_cb = mock.Mock()
        pipe.connect("died", pipeline_died_cb)

        with mock.patch.object(pipe, "set_state") as set_state:
            pipe.pause()
            set_state.assert_called_once_with(Gst.State.PAUSED)
        self.assertEqual(pipe._attempted_recoveries, 0)
        self.assertFalse(pipeline_died_cb.called)

        for i in range(MAX_RECOVERIES):
            with mock.patch.object(pipe, "set_state") as set_state:
                set_state.return_value = Gst.StateChangeReturn.SUCCESS
                self.post_fake_error_message(pipe)
                set_state.assert_has_calls([mock.call(Gst.State.NULL),
                                            mock.call(Gst.State.PAUSED)])
                self.assertEqual(set_state.call_count, 2)
            self.assertEqual(pipe._attempted_recoveries, i + 1)
            self.assertFalse(pipeline_died_cb.called)

        with mock.patch.object(pipe, "set_state") as set_state:
            self.post_fake_error_message(pipe)
            set_state.assert_not_called()
        self.assertTrue(pipeline_died_cb.called)
        self.assertEqual(pipe._attempted_recoveries, MAX_RECOVERIES)

    def test_async_done_not_received(self):
        """Checks the recovery when the ASYNC_DONE message timed out."""
        ges_timeline = GES.Timeline.new()
        self.assertTrue(ges_timeline.add_track(GES.VideoTrack.new()))
        ges_layer = ges_timeline.append_layer()
        uri = common.get_sample_uri("tears_of_steel.webm")
        asset = GES.UriClipAsset.request_sync(uri)
        ges_clip = asset.extract()
        self.assertTrue(ges_layer.add_clip(ges_clip))
        self.assertFalse(ges_timeline.is_empty())

        pipe = Pipeline(app=common.create_pitivi_mock())

        pipe.set_timeline(ges_timeline)
        self.assertFalse(pipe._busy_async)
        self.assertEqual(pipe._recovery_state, SimplePipeline.RecoveryState.NOT_RECOVERING)

        # Pretend waiting for async-done timed out.
        # We mock set_state because we don't actually care about the state,
        # and setting the state to PAUSED could show a video window.
        with mock.patch.object(pipe, "set_state"):
            pipe._async_done_not_received_cb("reason1", 1)
        # Make sure the pipeline started a watchdog timer waiting for async-done
        # as part of setting the state from NULL to PAUSED.
        self.assertTrue(pipe._busy_async)
        self.assertEqual(pipe._attempted_recoveries, 1)
        self.assertEqual(pipe._recovery_state, SimplePipeline.RecoveryState.STARTED_RECOVERING)

        # Pretend the state changed to READY.
        message = mock.Mock()
        message.type = Gst.MessageType.STATE_CHANGED
        message.src = pipe._pipeline
        message.parse_state_changed.return_value = (Gst.State.NULL, Gst.State.READY, Gst.State.PAUSED)
        pipe._bus_message_cb(None, message)

        # Pretend the state changed to PAUSED.
        message.parse_state_changed.return_value = (Gst.State.READY, Gst.State.PAUSED, Gst.State.VOID_PENDING)
        self.assertEqual(pipe._next_seek, None)
        pipe._bus_message_cb(None, message)
        self.assertEqual(pipe._recovery_state, SimplePipeline.RecoveryState.SEEKED_AFTER_RECOVERING)
        self.assertTrue(pipe._busy_async)
        # The pipeline should have tried to seek back to the last position.
        self.assertEqual(pipe._next_seek, 0)

        # Pretend the state change (to PAUSED) async operation succeeded.
        message.type = Gst.MessageType.ASYNC_DONE
        with mock.patch.object(pipe, "get_state") as get_state:
            get_state.return_value = (0, Gst.State.PAUSED, 0)
            pipe._bus_message_cb(None, message)
        self.assertEqual(pipe._recovery_state, SimplePipeline.RecoveryState.NOT_RECOVERING)
        # Should still be busy because of seeking to _next_seek.
        self.assertTrue(pipe._busy_async)
        self.assertIsNone(pipe._next_seek)

        # Pretend the seek async operation finished.
        message.type = Gst.MessageType.ASYNC_DONE
        pipe._bus_message_cb(None, message)
        self.assertEqual(pipe._recovery_state, SimplePipeline.RecoveryState.NOT_RECOVERING)
        self.assertFalse(pipe._busy_async)
        self.assertIsNone(pipe._next_seek)

    def test_commit_timeline_after(self):
        """Checks the recovery mechanism."""
        pipe = Pipeline(common.create_pitivi_mock())
        timeline = GES.Timeline()
        pipe.set_timeline(timeline)

        with mock.patch.object(pipe, "get_state") as get_state:
            get_state.return_value = (0, Gst.State.PAUSED, 0)
            with mock.patch.object(timeline, "commit") as commit:
                with pipe.commit_timeline_after():
                    pipe.commit_timeline()
                self.assertEqual(commit.call_count, 1)

            with mock.patch.object(timeline, "commit") as commit:
                with pipe.commit_timeline_after():
                    self.assertEqual(pipe._prevent_commits, 1)
                    with pipe.commit_timeline_after():
                        self.assertEqual(pipe._prevent_commits, 2)
                        pipe.commit_timeline()
                        self.assertEqual(commit.call_count, 0)
                self.assertEqual(commit.call_count, 1)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2012, Thibault Saunier <tsaunier@gnome.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Logic for undo/redo project actions."""
from gi.repository import GES
from gi.repository import GObject
from gi.repository import Gst

from pitivi.undo.timeline import TimelineObserver
from pitivi.undo.undo import Action
from pitivi.undo.undo import MetaContainerObserver
from pitivi.undo.undo import UndoableAction


class AssetAddedIntention(UndoableAction):
    """The intention of adding an asset to a project.

    This should be created when the async operation starts.
    See also AssetAddedAction.
    """

    def __init__(self, project, uri):
        UndoableAction.__init__(self)
        self.project = project
        self.uri = uri
        self.asset = None
        self.project.connect("asset-added", self._asset_added_cb)

    def _asset_added_cb(self, project, asset):
        if asset.get_id() == self.uri:
            self.asset = asset
            self.project.disconnect_by_func(self._asset_added_cb)

    def undo(self):
        # The asset might be missing if removed before it's added
        if self.asset:
            self.project.remove_asset(self.asset)

    def do(self):
        if self.asset:
            self.project.add_asset(self.asset)


class AssetAddedAction(Action):
    """The adding of an asset to a project.

    This should be created when the asset has been added.
    See also AssetAddedIntention.
    """

    def __init__(self, asset):
        Action.__init__(self)
        self.asset = asset

    def as_scenario_action(self):
        st = Gst.Structure.new_empty("add-asset")
        st.set_value("id", self.asset.get_id())
        type_string = GObject.type_name(self.asset.get_extractable_type())
        st.set_value("type", type_string)
        return st


class AssetRemovedAction(UndoableAction):
    """The removal of an asset from a project."""

    def __init__(self, project, asset):
        UndoableAction.__init__(self)
        self.project = project
        self.asset = asset

    def undo(self):
        self.project.add_asset(self.asset)

    def do(self):
        self.project.remove_asset(self.asset)

    def as_scenario_action(self):
        st = Gst.Structure.new_empty("remove-asset")
        st.set_value("id", self.asset.get_id())
        type_string = GObject.type_name(self.asset.get_extractable_type())
        st.set_value("type", type_string)
        return st


class AssetProxiedIntention(UndoableAction):
    """The intention of proxying an asset.

    Attributes:
        asset (GES.Asset): The original asset to be proxied.
        proxy_manager (pitivi.utils.proxy.ProxyManager): The manager
            controlling the proxy generation.
    """

    def __init__(self, asset, project, proxy_manager):
        UndoableAction.__init__(self)
        self.asset = asset
        self.project = project
        self.proxy_manager = proxy_manager

    def do(self):
        self.asset.force_proxying = True
        self.proxy_manager.add_job(self.asset)

    def undo(self):
        self.proxy_manager.cancel_job(self.asset)
        proxy = self.asset.props.proxy
        self.asset.set_proxy(None)
        if proxy:
            self.project.remove_asset(proxy)


class ProjectObserver(MetaContainerObserver):
    """Monitors a project instance and reports UndoableActions.

    Args:
        project (Project): The project to be monitored.
    """

    def __init__(self, project, action_log):
        MetaContainerObserver.__init__(self, project, action_log)
        project.connect("asset-added", self._asset_added_cb)
        project.connect("asset-removed", self._asset_removed_cb)
        assets = project.list_assets(GES.Extractable)
        for asset in assets:
            MetaContainerObserver.__init__(self, asset, action_log)
        self.timeline_observer = TimelineObserver(project.ges_timeline,
                                                  action_log)

    def _asset_added_cb(self, unused_project, asset):
        if not isinstance(asset, GES.UriClipAsset):
            return
        MetaContainerObserver.__init__(self, asset, self.action_log)
        action = AssetAddedAction(asset)
        self.action_log.push(action)

    def _asset_removed_cb(self, project, asset):
        if not isinstance(asset, GES.UriClipAsset):
            return
        action = AssetRemovedAction(project, asset)
        self.action_log.push(action)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Jaden Goter <jadengoter@huskers.unl.edu>
# Copyright (c) 2020, Jessie Guo <jessie.guo@huskers.unl.edu>
# Copyright (c) 2020, Daniel Rudebusch <daniel.rudebusch@huskers.unl.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the pitivi.viewer.guidelines module."""
# pylint: disable=protected-access,no-self-use,attribute-defined-outside-init
from pitivi.viewer.guidelines import Guideline
from pitivi.viewer.viewer import ViewerContainer
from tests import common


class GuidelinesPopoverTest(common.TestCase):
    """Tests for the guidelines classes."""

    def setup_viewer_widget(self):
        """Sets a viewer widget up."""
        self.app = common.create_pitivi()
        self.viewer_container = ViewerContainer(self.app)
        project = self.app.project_manager.new_blank_project()
        self.viewer_container.set_project(project)

    def _check_guidelines(self, guidelines, toggled_guidelines):
        overlay = self.viewer_container.guidelines_popover.overlay

        self.assertSetEqual(overlay.active_guidelines, set(guidelines))
        self.assertEqual(overlay.get_visible(), bool(guidelines), guidelines)

        self.viewer_container.toggle_guidelines_action.activate(None)
        self.assertSetEqual(overlay.active_guidelines, set(toggled_guidelines))
        self.assertEqual(overlay.get_visible(), bool(toggled_guidelines), guidelines)

        self.viewer_container.toggle_guidelines_action.activate(None)
        self.assertSetEqual(overlay.active_guidelines, set(guidelines))
        self.assertEqual(overlay.get_visible(), bool(guidelines), guidelines)

    def test_activate_deactivate_toggle(self):
        self.setup_viewer_widget()
        popover = self.viewer_container.guidelines_popover

        self._check_guidelines([], [Guideline.three_by_three])

        all_guidelines = set()
        for guideline in Guideline:
            popover.switches[guideline].set_active(True)
            self._check_guidelines([guideline], [])

            popover.switches[guideline].set_active(False)
            self._check_guidelines([], [guideline])

            all_guidelines.add(guideline)

        for guideline in Guideline:
            popover.switches[guideline].set_active(True)
        self._check_guidelines(all_guidelines, [])

        for guideline in Guideline:
            popover.switches[guideline].set_active(False)
            expected = set(all_guidelines)
            expected.remove(guideline)
            self._check_guidelines(expected, [])

            popover.switches[guideline].set_active(True)
            self._check_guidelines(all_guidelines, [])



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Pitivi's about dialog."""
from gettext import gettext as _

from gi.repository import GES
from gi.repository import Gst
from gi.repository import Gtk

from pitivi.configure import APPNAME
from pitivi.configure import APPURL
from pitivi.configure import VERSION


class AboutDialog(Gtk.AboutDialog):
    """Pitivi's about dialog.

    Displays info regarding Pitivi's version, license,
    maintainers, contributors, etc.

    Attributes:
        app (Pitivi): The app.
    """

    def __init__(self, app):
        Gtk.AboutDialog.__init__(self)
        self.set_program_name(APPNAME)
        self.set_website(APPURL)

        if app.is_latest():
            version_str = _("Version %s") % VERSION
        else:
            version_str = _("Version %(cur_ver)s — %(new_ver)s is available") % \
                {"cur_ver": VERSION,
                 "new_ver": app.get_latest()}
        self.set_version(version_str)

        comments = ["",
                    "GES %s" % ".".join(map(str, GES.version())),
                    "GTK+ %s" % ".".join(map(str, (Gtk.MAJOR_VERSION, Gtk.MINOR_VERSION))),
                    "GStreamer %s" % ".".join(map(str, Gst.version()))]
        self.set_comments("\n".join(comments))

        authors = [_("Current maintainers:"),
                   "Thibault Saunier <tsaunier@gnome.org>",
                   "Mathieu Duponchelle <mduponchelle1@gmail.com>",
                   "Alexandru Băluț <alexandru.balut@gmail.com>",
                   "",
                   _("Past maintainers:"),
                   "Edward Hervey <bilboed@bilboed.com>",
                   "Alessandro Decina <alessandro.decina@collabora.co.uk>",
                   "Brandon Lewis <brandon_lewis@berkeley.edu>",
                   "Jean-François Fortin Tam <nekohayo@gmail.com>",
                   "",
                   _("Contributors:"),
                   "https://gitlab.gnome.org/GNOME/pitivi/-/graphs/master",
                   ""]
        self.set_authors(authors)
        # Translators: See
        # https://developer.gnome.org/gtk3/stable/GtkAboutDialog.html#gtk-about-dialog-set-translator-credits
        # for details on how this is used.
        translators = _("translator-credits")
        if translators != "translator-credits":
            self.set_translator_credits(translators)
        documenters = ["Jean-François Fortin Tam <nekohayo@gmail.com>", ]
        self.set_documenters(documenters)
        self.set_license_type(Gtk.License.LGPL_2_1)
        self.set_logo_icon_name("org.pitivi.Pitivi")
        self.connect("response", self.__about_response_cb)
        self.set_transient_for(app.gui)

    @staticmethod
    def __about_response_cb(dialog, unused_response):
        dialog.destroy()



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2005, Edward Hervey <bilboed@bilboed.com>
# Copyright (c) 2011, Benjamin M. Schwartz <bens@alum.mit.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Classes for extracting decoded contents of streams into Python

Code derived from ui/previewer.py.
"""
# FIXME reimplement after GES port
from collections import deque

from gi.repository import Gst

from pitivi.utils.loggable import Loggable
# from pitivi.elements.singledecodebin import SingleDecodeBin
# from pitivi.elements.extractionsink import ExtractionSink


def linkDynamic(element, target):

    def pad_added(unused_bin, pad, target):
        compatpad = target.get_compatible_pad(pad)
        if compatpad:
            pad.link_full(compatpad, Gst.PAD_LINK_CHECK_NOTHING)
    element.connect("pad-added", pad_added, target)


def pipeline(graph):
    E = iter(graph.items())
    V = iter(graph.keys())
    p = Gst.Pipeline()
    p.add(*V)
    for u, v in E:
        if v:
            try:
                u.link(v)
            except Gst.LinkError:
                linkDynamic(u, v)
    return p


class Extractee:

    """Abstract base class for receiving raw data from an L{Extractor}."""

    def receive(self, array):
        """
        Receive a chunk of data from an Extractor.

        @param array: The chunk of data as an array
        @type array: any kind of numeric array

        """
        raise NotImplementedError

    def finalize(self):
        """
        Inform the Extractee that receive() will not be called again.

        Indicates that the extraction is complete, so the Extractee should
            process the data it has received.

        """
        raise NotImplementedError


class Extractor(Loggable):

    """
    Abstract base class for extraction of raw data from a stream.

    Closely modeled on L{Previewer}.

    """

    def __init__(self, factory, stream_):
        """
        Create a new Extractor.

        @param factory: the factory with which to decode the stream
        @type factory: L{ObjectFactory}
        @param stream_: the stream to decode
        @type stream_: L{Stream}
        """
        Loggable.__init__(self)
        self.debug("Initialized with %s %s", factory, stream_)

    def extract(self, extractee, start, duration):
        """
        Extract the raw data corresponding to a segment of the stream.

        @param extractee: the L{Extractee} that will receive the raw data
        @type extractee: L{Extractee}
        @param start: The point in the stream at which the segment starts
            (nanoseconds)
        @type start: L{long}
        @param duration: The duration of the segment (nanoseconds)
        @type duration: L{long}

        """
        raise NotImplementedError


class RandomAccessExtractor(Extractor):

    """
    Abstract class for L{Extractor}s of random access streams.

    Closely inspired by L{RandomAccessPreviewer}.

    """

    def __init__(self, factory, stream_):
        Extractor.__init__(self, factory, stream_)
        # FIXME:
        # why doesn't this work?
        # bin = factory.makeBin(stream_)
        uri = factory.uri
        caps = stream_.caps
        bin = SingleDecodeBin(uri=uri, caps=caps, stream=stream_)

        self._pipelineInit(factory, bin)

    def _pipelineInit(self, factory, bin):
        """
        Create the pipeline for the preview process.

        Subclasses should
        override this method and create a pipeline, connecting to
        callbacks to the appropriate signals, and prerolling the
        pipeline if necessary.

        """
        raise NotImplementedError


class RandomAccessAudioExtractor(RandomAccessExtractor):

    """
    L{Extractor} for random access audio streams.

    Closely inspired by L{RandomAccessAudioPreviewer}.

    """

    def __init__(self, factory, stream_):
        self._queue = deque()
        RandomAccessExtractor.__init__(self, factory, stream_)
        self._ready = False

    def _pipelineInit(self, factory, sbin):
        self.audioSink = ExtractionSink()
        self.audioSink.set_stopped_cb(self._finishSegment)
        # This audiorate element ensures that the extracted raw-data
        # timeline matches the timestamps used for seeking, even if the
        # audio source has gaps or other timestamp abnormalities.
        audiorate = Gst.ElementFactory.make("audiorate")
        conv = Gst.ElementFactory.make("audioconvert")
        q = Gst.ElementFactory.make("queue")
        self.audioPipeline = pipeline({
            sbin: audiorate,
            audiorate: conv,
            conv: q,
            q: self.audioSink,
            self.audioSink: None})
        bus = self.audioPipeline.get_bus()
        bus.add_signal_watch()
        bus.connect("message::error", self._busMessageErrorCb)
        self._donecb_id = bus.connect("message::async-done",
                                      self._busMessageAsyncDoneCb)

        self.audioPipeline.set_state(Gst.State.PAUSED)
        # The audiopipeline.set_state() method does not take effect
        # immediately, but the extraction process (and in particular
        # self._startSegment) will not work properly until
        # self.audioPipeline reaches the desired state (State.PAUSED).
        # To ensure that this is the case, we wait until the ASYNC_DONE
        # message is received before setting self._ready = True,
        # which enables extraction to proceed.

    def _busMessageErrorCb(self, unused_bus, message):
        error, debug = message.parse_error()
        self.error("Event bus error: %s; %s", error, debug)

        return Gst.BusSyncReply.PASS

    def _busMessageAsyncDoneCb(self, bus, unused_message):
        self.debug("Pipeline is ready for seeking")
        bus.disconnect(self._donecb_id)  # Don't call me again
        self._ready = True
        if self._queue:  # Someone called .extract() before we were ready
            self._run()

    def _startSegment(self, timestamp, duration):
        self.debug("processing segment with timestamp=%i and duration=%i",
                   timestamp, duration)
        res = self.audioPipeline.seek(1.0,
                                      Gst.Format.TIME,
                                      Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE,
                                      Gst.SeekType.SET, timestamp,
                                      Gst.SeekType.SET, timestamp + duration)
        if not res:
            self.warning("seek failed %s", timestamp)
        self.audioPipeline.set_state(Gst.State.PLAYING)

        return res

    def _finishSegment(self):
        self.audioSink.extractee.finalize()
        self.audioSink.reset()
        self._queue.popleft()
        # If there's more to do, keep running
        if self._queue:
            self._run()

    def extract(self, extractee, start, duration):
        stopped = not self._queue
        self._queue.append((extractee, start, duration))
        if stopped and self._ready:
            self._run()
        # if self._ready is False, self._run() will be called from
        # self._busMessageDoneCb().

    def _run(self):
        # Control flows in a cycle:
        # _run -> _startSegment -> busMessageSegmentDoneCb -> _finishSegment -> _run
        # This forms a loop that extracts an entire segment (i.e. satisfies an
        # extract request) in each cycle. The cycle
        # runs until the queue of Extractees empties.  If the cycle is not
        # running, extract() will kick it off again.
        extractee, start, duration = self._queue[0]
        self.audioSink.set_extractee(extractee)
        self._startSegment(start, duration)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Millan Castro <m.castrovilarino@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Undo/redo logic for markers."""
from gi.repository import Gst

from pitivi.undo.undo import MetaContainerObserver
from pitivi.undo.undo import UndoableAutomaticObjectAction
from pitivi.utils.loggable import Loggable


class MarkerListObserver(Loggable):
    """Monitors a MarkerList and reports UndoableActions.

    Args:
        ges_marker_list (GES.MarkerList): The markerlist to observe.

    Attributes:
        action_log (UndoableActionLog): The action log where to report actions.
    """

    def __init__(self, ges_marker_list, action_log):
        Loggable.__init__(self)

        self.action_log = action_log

        self.marker_observers = {}

        ges_marker_list.connect("marker-added", self._marker_added_cb)
        ges_marker_list.connect("marker-removed", self._marker_removed_cb)
        ges_marker_list.connect("marker-moved", self._marker_moved_cb)

        ges_markers = ges_marker_list.get_markers()
        for ges_marker in ges_markers:
            self._connect(ges_marker)

    def _connect(self, ges_marker):
        marker_observer = MetaContainerObserver(ges_marker, self.action_log)
        self.marker_observers[ges_marker] = marker_observer

    def _marker_added_cb(self, ges_marker_list, position, ges_marker):
        action = MarkerAdded(ges_marker_list, ges_marker)
        self.action_log.push(action)
        self._connect(ges_marker)

    def _marker_removed_cb(self, ges_marker_list, ges_marker):
        action = MarkerRemoved(ges_marker_list, ges_marker)
        self.action_log.push(action)
        marker_observer = self.marker_observers.pop(ges_marker)
        marker_observer.release()

    def _marker_moved_cb(self, ges_marker_list, old_position, position, ges_marker):
        action = MarkerMoved(ges_marker_list, ges_marker, old_position)
        self.action_log.push(action)


# pylint: disable=abstract-method
class MarkerAction(UndoableAutomaticObjectAction):
    """Base class for marker actions."""

    def __init__(self, ges_marker_list, ges_marker):
        UndoableAutomaticObjectAction.__init__(self, ges_marker)
        self.ges_marker_list = ges_marker_list
        self.position = ges_marker.props.position
        self.ges_marker = ges_marker

    def add(self):
        """Adds a marker and updates the auto-object."""
        ges_marker = self.ges_marker_list.add(self.position)
        comment = self.auto_object.get_string("comment")
        if comment:
            ges_marker.set_string("comment", comment)
        UndoableAutomaticObjectAction.update_object(self.auto_object, ges_marker)

    def remove(self):
        """Removes the marker represented by the auto_object."""
        self.ges_marker_list.remove(self.auto_object)


class MarkerAdded(MarkerAction):
    """Action for added markers."""

    def do(self):
        self.add()

    def undo(self):
        self.remove()

    def as_scenario_action(self):
        st = Gst.Structure.new_empty("add-marker")
        return st


class MarkerRemoved(MarkerAction):
    """Action for removed markers."""

    def do(self):
        self.remove()

    def undo(self):
        self.add()

    def as_scenario_action(self):
        st = Gst.Structure.new_empty("remove-marker")
        return st


class MarkerMoved(UndoableAutomaticObjectAction):
    """Action for moved markers."""

    def __init__(self, ges_marker_list, ges_marker, old_position):
        UndoableAutomaticObjectAction.__init__(self, ges_marker)
        self.ges_marker_list = ges_marker_list
        self.new_position = ges_marker.props.position
        self.old_position = old_position

    def do(self):
        self.ges_marker_list.move(self.auto_object, self.new_position)

    def undo(self):
        self.ges_marker_list.move(self.auto_object, self.old_position)

    def as_scenario_action(self):
        st = Gst.Structure.new_empty("move-marker")
        return st

    def expand(self, action):
        if not isinstance(action, MarkerMoved):
            return False

        self.new_position = action.new_position
        return True



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2016, Lubosz Sarnecki <lubosz.sarnecki@collabora.co.uk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import cairo
import numpy

from pitivi.viewer.overlay import Overlay


class TitleOverlay(Overlay):
    """Viewer overlays for GES.TitleSource."""

    def __init__(self, stack, source):
        Overlay.__init__(self, stack, source)
        self.__position = numpy.array([0, 0])
        self.__size = None
        self.__click_source_position = None
        self.__click_window_position = None
        self.update_from_source()

        stack.app.project_manager.current_project.pipeline.connect("async-done", self.on_async_done)

    def on_async_done(self, unused_pipeline):
        # Only update on_async when we are not dragging
        if isinstance(self.stack.click_position, numpy.ndarray):
            return
        self.update_from_source()

    def __draw_rectangle(self, cr):
        x, y = [int(v) + 0.5 for v in self.__position]
        w, h = [int(v) - 1 for v in self.__size]
        cr.rectangle(x, y, w, h)

    def __get_source_position(self):
        res_x, x = self._source.get_child_property("x-absolute")
        res_y, y = self._source.get_child_property("y-absolute")
        assert res_x and res_y
        return numpy.array([x, y])

    def __get_text_position(self):
        res_x, x = self._source.get_child_property("text-x")
        res_y, y = self._source.get_child_property("text-y")
        assert res_x and res_y
        return numpy.array([x, y])

    def __get_text_size(self):
        res_w, w = self._source.get_child_property("text-width")
        res_h, h = self._source.get_child_property("text-height")
        assert res_w and res_h
        return numpy.array([w, h])

    def __set_source_position(self, position):
        self._source.set_child_property("x-absolute", float(position[0]))
        self._source.set_child_property("y-absolute", float(position[1]))

    def __update_from_motion(self, title_position):
        self.__position = title_position

    def update_from_source(self):
        position = self.__get_text_position()
        size = self.__get_text_size()

        self.__position = position * self.stack.window_size / self.project_size
        self.__size = size * self.stack.window_size / self.project_size
        self.queue_draw()

    def on_hover(self, cursor_position):
        if (self.__position < cursor_position).all() and (cursor_position < self.__position + self.__size).all():
            if self._is_selected():
                self.stack.set_cursor("grab")
            self._hover()
        else:
            self.unhover()
        self.queue_draw()
        return self._is_hovered()

    def on_button_press(self):
        self.__click_source_position = self.__get_source_position()
        self.__click_window_position = self.__position
        if self._is_hovered():
            self._select()
            self.stack.set_cursor("grabbing")
            self.stack.selected_overlay = self
        elif self._is_selected():
            self._deselect()

    def on_button_release(self, cursor_position):
        self.__click_source_position = None
        self.on_hover(cursor_position)
        if self._is_hovered():
            self.stack.set_cursor("grab")
        self.queue_draw()

    def on_motion_notify(self, cursor_position):
        if not isinstance(self.stack.click_position, numpy.ndarray):
            return

        self.__update_from_motion(self.__click_window_position + self.stack.get_drag_distance(cursor_position))
        self.queue_draw()

        title_position = self.__position / (self.stack.window_size * (1 - self.__get_text_size() / self.project_size))

        self.__set_source_position(title_position)
        self._commit()

    def do_draw(self, cr):
        selected = self._is_selected()
        hovered = self._is_hovered()
        if not selected and not hovered:
            return

        cr.save()

        # Clear background
        cr.set_operator(cairo.OPERATOR_OVER)
        cr.set_source_rgba(0, 0, 0, 0)
        cr.paint()

        if not selected:
            cr.set_dash((5, 5))

        # Black outline around the box
        cr.set_source_rgb(0, 0, 0)
        cr.set_line_width(3)
        self.__draw_rectangle(cr)
        cr.stroke()

        # Inner white line
        color = (0.8, 0.8, 0.8) if not selected else (1, 1, 1)
        cr.set_source_rgb(*color)
        cr.set_line_width(1)
        self.__draw_rectangle(cr)
        cr.stroke()

        cr.restore()



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Millan Castro <m.castrovilarino@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the undo.markers module."""
import tempfile

from gi.repository import Gst

from tests import common


class TestMarkers(common.TestCase):
    """Tests for the various classes."""

    @common.setup_timeline
    def test_marker_added(self):
        """Checks marker addition."""
        markers = self.timeline.get_marker_list("markers")

        with self.action_log.started("Added marker"):
            marker1 = markers.add(10)
        self.assert_markers(markers, [(10, None)])

        with self.action_log.started("new comment"):
            marker1.set_string("comment", "comment 1")
        self.assert_markers(markers, [(10, "comment 1")])

        for _ in range(4):
            self.action_log.undo()
            self.assert_markers(markers, [(10, None)])

            self.action_log.undo()
            self.assert_markers(markers, [])

            self.action_log.redo()
            self.assert_markers(markers, [(10, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(10, "comment 1")])

        with self.action_log.started("Added marker"):
            marker2 = markers.add(20)
        self.assert_markers(markers, [(10, "comment 1"), (20, None)])

        with self.action_log.started("new comment"):
            marker2.set_string("comment", "comment 2")
        self.assert_markers(markers, [(10, "comment 1"), (20, "comment 2")])

        for _ in range(4):
            self.action_log.undo()
            self.action_log.undo()
            self.action_log.undo()
            self.action_log.undo()
            self.assert_markers(markers, [])

            self.action_log.redo()
            self.assert_markers(markers, [(10, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(10, "comment 1")])

            self.action_log.redo()
            self.assert_markers(markers, [(10, "comment 1"), (20, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(10, "comment 1"), (20, "comment 2")])

    @common.setup_timeline
    def test_marker_removed(self):
        """Checks marker removal."""
        markers = self.timeline.get_marker_list("markers")
        marker1 = markers.add(10)
        marker2 = markers.add(20)

        with self.action_log.started("Removed marker"):
            markers.remove(marker1)
        self.assert_markers(markers, [(20, None)])

        with self.action_log.started("Removed marker"):
            markers.remove(marker2)
        self.assert_markers(markers, [])

        for _ in range(4):
            self.action_log.undo()
            self.assert_markers(markers, [(20, None)])

            self.action_log.undo()
            self.assert_markers(markers, [(10, None), (20, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(20, None)])

            self.action_log.redo()
            self.assert_markers(markers, [])

    @common.setup_timeline
    def test_marker_moved(self):
        """Checks marker moving."""
        markers = self.timeline.get_marker_list("markers")
        marker1 = markers.add(10)
        markers.add(20)

        with self.action_log.started("Move marker"):
            markers.move(marker1, 40)
            markers.move(marker1, 30)

        stack, = self.action_log.undo_stacks
        self.assertEqual(len(stack.done_actions), 1, stack.done_actions)

        self.assert_markers(markers, [(20, None), (30, None)])

        for _ in range(4):
            self.action_log.undo()
            self.assert_markers(markers, [(10, None), (20, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(20, None), (30, None)])

    @common.setup_timeline
    def test_marker_comment(self):
        """Checks marker comment."""
        markers = self.timeline.get_marker_list("markers")

        with self.action_log.started("Added marker"):
            marker1 = markers.add(10)
        with self.action_log.started("Added marker"):
            marker2 = markers.add(20)

        self.assert_markers(markers, [(10, None), (20, None)])

        with self.action_log.started("new comment"):
            marker1.set_string("comment", "comment 1")
        with self.action_log.started("new comment"):
            marker2.set_string("comment", "comment 2")

        self.assert_markers(markers, [(10, "comment 1"), (20, "comment 2")])

        for _ in range(4):
            self.action_log.undo()
            self.assert_markers(markers, [(10, "comment 1"), (20, None)])

            self.action_log.undo()
            self.assert_markers(markers, [(10, None), (20, None)])

            self.action_log.undo()
            self.assert_markers(markers, [(10, None)])

            self.action_log.undo()
            self.assert_markers(markers, [])

            self.action_log.redo()
            self.assert_markers(markers, [(10, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(10, None), (20, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(10, "comment 1"), (20, None)])

            self.action_log.redo()
            self.assert_markers(markers, [(10, "comment 1"), (20, "comment 2")])

    @common.setup_timeline
    def test_marker_load_project(self):
        """Checks marker addition."""
        # TODO: When there is nothing connected to closing-project,
        # the default reply is "False", which means "abort saving". It should mean
        # "OK" to get rid off the handler.  The meaning of the default (False)
        # must be changed
        def closing(unused_manager, unused_project):
            return True

        def loaded_cb(project, timeline):
            mainloop.quit()

        markers = self.timeline.get_marker_list("markers")
        markers.add(10)
        markers.add(20)
        self.assert_markers(markers, [(10, None), (20, None)])

        project_uri = Gst.filename_to_uri(tempfile.NamedTemporaryFile().name)
        self.app.project_manager.save_project(project_uri)
        self.app.project_manager.connect("closing-project", closing)

        self.app.project_manager.close_running_project()
        project = self.app.project_manager.new_blank_project()
        markers = project.ges_timeline.get_marker_list("markers")
        self.assert_markers(markers, [])

        self.app.project_manager.close_running_project()
        project = self.app.project_manager.load_project(project_uri)
        project.connect("loaded", loaded_cb)
        mainloop = common.create_main_loop()
        mainloop.run()
        action_log = self.app.action_log

        markers = project.ges_timeline.get_marker_list("markers")
        self.assert_markers(markers, [(10, None), (20, None)])

        ges_markers = markers.get_markers()
        marker1, marker2 = ges_markers

        with action_log.started("new comment"):
            marker1.set_string("comment", "comment 1")
        self.assert_markers(markers, [(10, "comment 1"), (20, None)])

        with action_log.started("new comment"):
            marker2.set_string("comment", "comment 2")
        self.assert_markers(markers, [(10, "comment 1"), (20, "comment 2")])

        for _ in range(4):
            action_log.undo()
            self.assert_markers(markers, [(10, "comment 1"), (20, None)])

            action_log.undo()
            self.assert_markers(markers, [(10, None), (20, None)])

            action_log.redo()
            self.assert_markers(markers, [(10, "comment 1"), (20, None)])

            action_log.redo()
            self.assert_markers(markers, [(10, "comment 1"), (20, "comment 2")])



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2017, Fabian Orccon <cfoch.fabian@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the pluginmanager module."""
import os
import tempfile
from unittest import mock

from gi.repository import GObject

from pitivi.pluginmanager import PluginManager
from pitivi.settings import GlobalSettings
from tests import common


class TestPluginManager(common.TestCase):
    """Tests for the PluginManager class."""

    def test_load_plugins_from_settings(self):
        """Checks if the plugin manager loads plugins from GlobalSettings."""
        class App(GObject.Object):
            """A representation of the Pitivi Application for test purposes."""

            __gsignals__ = {
                "window-added": (GObject.SignalFlags.RUN_LAST, None, (object, ))
            }

            def __init__(self):
                GObject.Object.__init__(self)
                self.settings = GlobalSettings()

        with mock.patch("pitivi.pluginmanager.get_plugins_dir") as get_plugins_dir,\
                mock.patch("pitivi.pluginmanager.get_user_plugins_dir") as get_user_plugins_dir,\
                tempfile.TemporaryDirectory() as temp_dir:

            plugin_content = ("[Plugin]\n"
                              "Module=pluginA\n"
                              "Name=PluginA\n"
                              "Loader=Python3")

            py_content = ("from gi.repository import GObject\n"
                          "class PluginA(GObject.GObject):\n"
                          "    def __init__(self):\n"
                          "        GObject.Object.__init__(self)")

            with open(os.path.join(temp_dir, "pluginA.plugin"), "w") as plugin_file:
                plugin_file.write(plugin_content)
            with open(os.path.join(temp_dir, "pluginA.py"), "w") as py_file:
                py_file.write(py_content)

            get_plugins_dir.return_value = temp_dir
            get_user_plugins_dir.return_value = temp_dir

            app = App()
            app.settings.ActivePlugins = ["pluginA"]

            plugin_manager = PluginManager(app)
            app.emit("window-added", None)

            loaded_plugins = plugin_manager.engine.get_loaded_plugins()
            self.assertCountEqual(loaded_plugins, app.settings.ActivePlugins)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2009, Alessandro Decina <alessandro.d@gmail.com>
# Copyright (c) 2013, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the pitivi.project module."""
# pylint: disable=protected-access
import collections
import os
import tempfile
import time
from unittest import mock

from gi.repository import GES
from gi.repository import Gst

from pitivi.medialibrary import AssetThumbnail
from pitivi.medialibrary import MediaLibraryWidget
from pitivi.project import Project
from pitivi.project import ProjectManager
from pitivi.utils.misc import path_from_uri
from pitivi.utils.proxy import ProxyingStrategy
from tests import common


class ProjectManagerListener():

    def __init__(self, manager):
        self.manager = manager
        self.connect_to_project_manager()
        self._reset()

    def _reset(self):
        self.signals = []

    def connect_to_project_manager(self):
        for signal in ("new-project-loading", "new-project-loaded",
                       "new-project-created", "new-project-failed", "missing-uri",
                       "closing-project", "project-closed"):
            self.manager.connect(signal, self._record_signal, signal)

    def _record_signal(self, *args):
        signal = args[-1]
        args = args[1:-1]
        self.signals.append((signal, args))

        return True


class TestProjectManager(common.TestCase):

    def setUp(self):
        super().setUp()
        self.setup_app()

    def setup_app(self, app=None):
        if not app:
            app = mock.MagicMock()
        self.manager = ProjectManager(app)
        self.listener = ProjectManagerListener(self.manager)
        self.signals = self.listener.signals

    def test_loading_missing_project_file(self):
        """Checks new-project-failed is emitted for unsuitable formatters."""
        project_uri = Gst.filename_to_uri(tempfile.NamedTemporaryFile().name)
        self.manager.load_project(project_uri)

        # loading
        name, args = self.signals[0]
        self.assertEqual(project_uri, args[0].get_uri(), self.signals)

        # failed
        name, args = self.signals[1]
        self.assertEqual("new-project-failed", name)
        signal_uri, unused_message = args
        self.assertEqual(project_uri, signal_uri, self.signals)

    def test_new_blank_project_signals(self):
        self.manager.new_blank_project()

        name, _args = self.signals[0]
        self.assertEqual("new-project-loading", name, self.signals)

        name, _args = self.signals[1]
        self.assertEqual("new-project-created", name, self.signals)

        name, _args = self.signals[2]
        self.assertEqual("new-project-loaded", name, self.signals)

    def test_missing_uri_forwarded(self):
        self.setup_app(app=common.create_pitivi_mock())
        mainloop = common.create_main_loop()

        def missing_uri_cb(self, project, error, clip_asset, result):
            result[0] = True
            mainloop.quit()

        result = [False]
        self.manager.connect("missing-uri", missing_uri_cb, result)

        with common.cloned_sample():
            asset_uri = common.get_sample_uri("missing.png")
            with common.created_project_file(asset_uri) as uri:
                self.assertIsNotNone(self.manager.load_project(uri))
                mainloop.run()
        self.assertTrue(result[0], "missing-uri has not been emitted")

    def test_loaded(self):
        mainloop = common.create_main_loop()

        def new_project_loaded_cb(project_manager, project):
            mainloop.quit()

        self.manager.connect("new-project-loaded", new_project_loaded_cb)

        with common.cloned_sample("flat_colour1_640x480.png"):
            asset_uri = common.get_sample_uri("flat_colour1_640x480.png")
            with common.created_project_file(asset_uri=asset_uri) as uri:
                project = self.manager.load_project(uri)
                self.assertIsNotNone(project)
                mainloop.run()

        self.assertFalse(project.at_least_one_asset_missing)
        self.assertTrue(project.loaded)
        self.assertFalse(project.has_unsaved_modifications())

    def test_close_running_project_no_project(self):
        self.assertTrue(self.manager.close_running_project())
        self.assertFalse(self.signals)

    def test_close_running_project_refuse_from_signal(self):
        def closing_cb(manager, project):
            return False

        self.manager.current_project = mock.Mock()
        self.manager.current_project.uri = "file:///ciao"
        self.manager.connect("closing-project", closing_cb)

        self.assertFalse(self.manager.close_running_project())
        self.assertEqual(1, len(self.signals))
        name, args = self.signals[0]
        self.assertEqual("closing-project", name)
        project = args[0]
        self.assertTrue(project is self.manager.current_project)

    def test_close_running_project(self):
        project = self.manager.new_blank_project()
        self.assertTrue(self.manager.close_running_project())
        self.assertEqual(5, len(self.signals), self.signals)

        name, args = self.signals[-2]
        self.assertEqual("closing-project", name)
        self.assertEqual(args[0], project)

        name, args = self.signals[-1]
        self.assertEqual("project-closed", name)
        self.assertEqual(args[0], project)

        self.assertTrue(self.manager.current_project is None)

    def test_new_blank_project(self):
        self.assertIsNotNone(self.manager.new_blank_project())
        self.assertEqual(3, len(self.signals))

        name, args = self.signals[0]
        self.assertEqual("new-project-loading", name)
        project = args[0]
        self.assertTrue(project.get_uri() is None)

        name, args = self.signals[1]
        self.assertEqual("new-project-created", name)
        project = args[0]
        self.assertEqual(project.get_uri(), project.uri)

        name, args = self.signals[2]
        self.assertEqual("new-project-loaded", name)
        project = args[0]
        self.assertTrue(project is self.manager.current_project)

    def test_marker_container(self):
        project = self.manager.new_blank_project()
        self.assertIsNotNone(project)
        self.assertIsNotNone(project.ges_timeline)
        self.assertIsNotNone(project.ges_timeline.get_marker_list("markers"))

    def test_save_project(self):
        self.manager.new_blank_project()

        unused, path = tempfile.mkstemp(suffix=".xges")
        unused, path2 = tempfile.mkstemp(suffix=".xges")
        try:
            uri = "file://" + os.path.abspath(path)
            uri2 = "file://" + os.path.abspath(path2)

            # Save the project.
            self.assertTrue(self.manager.save_project(uri=uri, backup=False))
            self.assertTrue(os.path.isfile(path))

            # Wait a bit.
            time.sleep(0.1)

            # Save the project at a new location.
            self.assertTrue(self.manager.save_project(uri2, backup=False))
            self.assertTrue(os.path.isfile(path2))

            # Make sure the old path and the new path have different mtimes.
            mtime = os.path.getmtime(path)
            mtime2 = os.path.getmtime(path2)
            self.assertLess(mtime, mtime2)

            # Wait a bit more.
            time.sleep(0.1)

            # Save project again under the new path (by omitting uri arg)
            self.assertTrue(self.manager.save_project(backup=False))

            # regression test for bug 594396
            # make sure we didn't save to the old URI
            self.assertEqual(mtime, os.path.getmtime(path))
            # make sure we did save to the new URI
            self.assertLess(mtime2, os.path.getmtime(path2))
        finally:
            os.remove(path)
            os.remove(path2)

    def test_make_backup_uri(self):
        uri = "file:///tmp/x.xges"
        self.assertEqual(uri + "~", self.manager._make_backup_uri(uri))

    def test_backup_project(self):
        self.manager.new_blank_project()

        # Assign an uri to the project where it's saved by default.
        unused, xges_path = tempfile.mkstemp(suffix=".xges")
        uri = "file://" + os.path.abspath(xges_path)
        self.manager.current_project.uri = uri
        # This is where the automatic backup file is saved.
        backup_uri = self.manager._make_backup_uri(uri)

        # Save the backup
        self.assertTrue(self.manager.save_project(
            self.manager.current_project, backup=True))
        self.assertTrue(os.path.isfile(path_from_uri(backup_uri)))

        self.manager.close_running_project()
        self.assertFalse(os.path.isfile(path_from_uri(backup_uri)),
                         "Backup file not deleted when project closed")


class TestProjectLoading(common.TestCase):

    def test_loaded_callback(self):
        mainloop = common.create_main_loop()

        def loaded_cb(project, timeline):
            # If not called, the timeout of the mainloop will fail the test.
            mainloop.quit()

        # Create a blank project and save it.
        project = common.create_project()
        project.connect("loaded", loaded_cb)
        mainloop.run()
        # The blank project loading succeeded emitting signal "loaded".

        self.assertIsNotNone(project.ges_timeline)
        self.assertEqual(len(project.ges_timeline.get_layers()), 1)

        # Load the blank project and make sure "loaded" is triggered.
        unused, xges_path = tempfile.mkstemp()
        uri = "file://%s" % xges_path
        try:
            # Save so we can close it without complaints.
            project.save(project.ges_timeline, uri, None, overwrite=True)

            project2 = common.create_project()
            project2.connect("loaded", loaded_cb)
            mainloop.run()
            # The blank project loading succeeded emitting signal "loaded".

            self.assertIsNotNone(project2.ges_timeline)
            self.assertEqual(len(project2.ges_timeline.get_layers()), 1)
        finally:
            os.remove(xges_path)

    def test_asset_added_signal(self):
        app = common.create_pitivi()
        project = app.project_manager.new_blank_project()
        proxy_manager = app.proxy_manager

        mainloop = common.create_main_loop()

        def asset_added_cb(project, asset, assets):
            assets.append(asset)

        assets = []
        project.connect("asset-added", asset_added_cb, assets)

        def proxy_ready_cb(unused_proxy_manager, asset, proxy):
            mainloop.quit()

        proxy_manager.connect("proxy-ready", proxy_ready_cb)

        uris = [common.get_sample_uri("tears_of_steel.webm")]
        project.add_uris(uris)

        mainloop.run()

        self.assertEqual(len(assets), 1, assets)

    def load_project_with_missing_proxy(self):
        """Loads a project with missing proxies."""
        uris = [common.get_sample_uri("1sec_simpsons_trailer.mp4")]
        proxy_uri = uris[0] + ".232417.proxy.mov"
        xges = r"""<ges version='0.3'>
  <project properties='properties;' metadatas='metadatas, name=(string)&quot;New\ Project&quot;, author=(string)Unknown, render-scale=(double)100;'>
    <encoding-profiles>
    </encoding-profiles>
    <ressources>
      <asset id='%(uri)s' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)6, duration=(guint64)1228000000;' metadatas='metadatas, audio-codec=(string)&quot;MPEG-4\ AAC\ audio&quot;, maximum-bitrate=(uint)130625, bitrate=(uint)130625, datetime=(datetime)2007-02-19T05:03:04Z, encoder=(string)Lavf54.6.100, container-format=(string)&quot;ISO\ MP4/M4A&quot;, video-codec=(string)&quot;H.264\ /\ AVC&quot;, file-size=(guint64)232417;'  proxy-id='file:///home/thiblahute/devel/pitivi/flatpak/pitivi/tests/samples/1sec_simpsons_trailer.mp4.232417.proxy.mov' />
      <asset id='%(proxy_uri)s' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)6, duration=(guint64)1228020833;' metadatas='metadatas, container-format=(string)Matroska, audio-codec=(string)Opus, language-code=(string)en, encoder=(string)Lavf54.6.100, bitrate=(uint)64000, video-codec=(string)&quot;Motion\ JPEG&quot;, file-size=(guint64)4695434;' />
    </ressources>
    <timeline properties='properties, auto-transition=(boolean)true, snapping-distance=(guint64)0;' metadatas='metadatas, duration=(guint64)0;'>
      <track caps='video/x-raw(ANY)' track-type='4' track-id='0' properties='properties, async-handling=(boolean)false, message-forward=(boolean)true, caps=(string)&quot;video/x-raw\(ANY\)&quot;, restriction-caps=(string)&quot;video/x-raw\,\ width\=\(int\)720\,\ height\=\(int\)576\,\ framerate\=\(fraction\)25/1&quot;, mixing=(boolean)true;' metadatas='metadatas;'/>
      <track caps='audio/x-raw(ANY)' track-type='2' track-id='1' properties='properties, async-handling=(boolean)false, message-forward=(boolean)true, caps=(string)&quot;audio/x-raw\(ANY\)&quot;, restriction-caps=(string)&quot;audio/x-raw\,\ format\=\(string\)S32LE\,\ channels\=\(int\)2\,\ rate\=\(int\)44100\,\ layout\=\(string\)interleaved&quot;, mixing=(boolean)true;' metadatas='metadatas;'/>
      <layer priority='0' properties='properties, auto-transition=(boolean)true;' metadatas='metadatas, volume=(float)1;'>
        <clip id='0' asset-id='%(proxy_uri)s' type-name='GESUriClip' layer-priority='0' track-types='6' start='0' duration='1228000000' inpoint='0' rate='0' properties='properties, name=(string)uriclip0, mute=(boolean)false, is-image=(boolean)false;' >
          <source track-id='1' children-properties='properties, GstVolume::mute=(boolean)false, GstVolume::volume=(double)1;'>
            <binding type='direct' source_type='interpolation' property='volume' mode='1' track_id='1' values =' 0:0.10000000000000001  1228000000:0.10000000000000001 '/>
          </source>
          <source track-id='0' children-properties='properties, GstFramePositioner::alpha=(double)1, GstDeinterlace::fields=(int)0, GstFramePositioner::height=(int)720, GstDeinterlace::mode=(int)0, GstFramePositioner::posx=(int)0, GstFramePositioner::posy=(int)0, GstDeinterlace::tff=(int)0, GstFramePositioner::width=(int)1280;'>
            <binding type='direct' source_type='interpolation' property='alpha' mode='1' track_id='0' values =' 0:1  1228000000:1 '/>
          </source>
        </clip>
      </layer>
      <groups>
      </groups>
    </timeline>
</project>
</ges>""" % {"uri": uris[0], "proxy_uri": proxy_uri}
        app = common.create_pitivi(proxying_strategy=ProxyingStrategy.ALL)
        app.recent_manager.remove_item = mock.Mock(return_value=True)
        proxy_manager = app.proxy_manager
        project_manager = app.project_manager
        medialib = MediaLibraryWidget(app)

        mainloop = common.create_main_loop()

        proj_uri = self.create_project_file_from_xges(xges)

        def closing_project_cb(*args, **kwargs):
            # Do not ask whether to save project on closing.
            return True

        def proxy_ready_cb(proxy_manager, asset, proxy):
            self.assertEqual(proxy.props.id, proxy_uri)
            mainloop.quit()

        project_manager.connect("closing-project", closing_project_cb)
        proxy_manager.connect_after("proxy-ready", proxy_ready_cb)

        app.project_manager.load_project(proj_uri)
        return mainloop, app, medialib, proxy_uri

    def test_load_project_with_missing_proxy(self):
        """Checks loading a project with missing proxies."""
        with common.cloned_sample("1sec_simpsons_trailer.mp4"):
            mainloop, _app, medialib, proxy_uri = self.load_project_with_missing_proxy()
            mainloop.run()

        self.assertEqual(medialib.store.get_n_items(), 1)
        self.assertEqual(medialib.store[0].asset.props.id,
                         proxy_uri)
        self.assertEqual(medialib.store[0].thumb_decorator.state,
                         AssetThumbnail.PROXIED)

    def test_load_project_with_missing_proxy_progress_tracking(self):
        """Checks progress tracking of loading project with missing proxies."""
        from gi.repository import GstTranscoder

        with common.cloned_sample("1sec_simpsons_trailer.mp4"):
            # Disable proxy generation by not making it start ever.
            # This way we are sure it will not finish before we test
            # the state while it is being rebuilt.
            with mock.patch.object(GstTranscoder.Transcoder, "run_async"):
                mainloop, app, medialib, _proxy_uri = self.load_project_with_missing_proxy()
                uri = common.get_sample_uri("1sec_simpsons_trailer.mp4")

                app.project_manager.connect("new-project-loaded", lambda x, y: mainloop.quit())
                mainloop.run()

        self.assertEqual(medialib.store.get_n_items(), 1)
        self.assertEqual(medialib.store[0].asset.props.id,
                         uri)
        self.assertEqual(medialib.store[0].thumb_decorator.state,
                         AssetThumbnail.IN_PROGRESS)

    def test_load_project_with_missing_proxy_stop_generating_and_proxy(self):
        """Checks cancelling creation of a missing proxies and forcing it again."""
        from gi.repository import GstTranscoder

        with common.cloned_sample("1sec_simpsons_trailer.mp4"):
            # Disable proxy generation by not making it start ever.
            # This way we are sure it will not finish before we test
            # stop generating the proxy and restart it.
            with mock.patch.object(GstTranscoder.Transcoder, "run_async"):
                mainloop, app, medialib, proxy_uri = self.load_project_with_missing_proxy()

                app.project_manager.connect("new-project-loaded", lambda x, y: mainloop.quit())
                mainloop.run()
                asset = medialib.store[0].asset
                app.project_manager.current_project.disable_proxies_for_assets([asset])

            row = medialib.store[0]
            asset = row.asset
            self.assertEqual(medialib._progressbar.get_fraction(), 1.0)
            uri = common.get_sample_uri("1sec_simpsons_trailer.mp4")
            self.assertEqual(asset.props.id, uri)
            self.assertEqual(asset.ready, True)
            self.assertEqual(asset.creation_progress, 100)
            self.assertEqual(row.thumb_decorator.state,
                             AssetThumbnail.NO_PROXY)

            app.project_manager.current_project.use_proxies_for_assets([asset])
            mainloop.run()

        row = medialib.store[0]
        asset = row.asset
        self.assertEqual(medialib._progressbar.is_visible(), False)
        self.assertEqual(asset.props.id, proxy_uri)
        self.assertEqual(asset.ready, True)
        self.assertEqual(asset.creation_progress, 100)
        self.assertEqual(row.thumb_decorator.state,
                         AssetThumbnail.PROXIED)

    def test_loading_project_with_moved_asset(self):
        """Loads a project with moved asset."""
        app = common.create_pitivi(proxying_strategy=ProxyingStrategy.NOTHING)

        proj_uri = self.create_project_file_from_xges("""<ges version='0.3'>
            <project properties='properties;' metadatas='metadatas;'>
                <ressources>
                    <asset id='file://this/is/a/moved/asset.mp4' extractable-type-name='GESUriClip'
                        properties='properties, supported-formats=(int)6, duration=(guint64)1228000000;' metadatas='metadatas' />
                </ressources>
            </project>
            </ges>""")
        project_manager = app.project_manager
        medialib = MediaLibraryWidget(app)

        mainloop = common.create_main_loop()

        def new_project_loaded_cb(*args, **kwargs):
            mainloop.quit()

        def missing_uri_cb(project_manager, project, unused_error, asset):
            return common.get_sample_uri("1sec_simpsons_trailer.mp4")

        project_manager.connect("missing-uri", missing_uri_cb)
        project_manager.connect("new-project-loaded", new_project_loaded_cb)

        project_manager.load_project(proj_uri)
        with common.cloned_sample("1sec_simpsons_trailer.mp4"):
            mainloop.run()
        self.assertEqual(medialib._progressbar.get_fraction(), 1.0)

    def test_loading_project_with_moved_assets_and_deleted_proxy(self):
        """Loads a project with moved asset and deleted proxy file."""
        mainloop = common.create_main_loop()

        created_proxies = []

        def proxy_ready_cb(unused_proxy_manager, asset, proxy):
            created_proxies.append(asset)
            if len(created_proxies) == 2:
                mainloop.quit()

        app = common.create_pitivi(proxying_strategy=ProxyingStrategy.ALL)
        app.proxy_manager.connect("proxy-ready", proxy_ready_cb)

        proj_uri = self.create_project_file_from_xges(r"""<ges version='0.3'>
  <project properties='properties;' metadatas='metadatas, name=(string)&quot;New\ Project&quot;, author=(string)Unknown, render-scale=(double)100, format-version=(string)0.3;'>
    <ressources>
      <asset id='file:///nop/1sec_simpsons_trailer.mp4' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)6, duration=(guint64)1228000000;' metadatas='metadatas, audio-codec=(string)&quot;MPEG-4\ AAC\ audio&quot;, maximum-bitrate=(uint)130625, bitrate=(uint)130625, datetime=(datetime)2007-02-19T05:03:04Z, encoder=(string)Lavf54.6.100, container-format=(string)&quot;ISO\ MP4/M4A&quot;, video-codec=(string)&quot;H.264\ /\ AVC&quot;, file-size=(guint64)232417;'  proxy-id='file:///nop/1sec_simpsons_trailer.mp4.232417.proxy.mov' />
      <asset id='file:///nop/tears_of_steel.webm' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)6, duration=(guint64)2003000000;' metadatas='metadatas, container-format=(string)Matroska, language-code=(string)und, application-name=(string)Lavc56.60.100, encoder=(string)&quot;Xiph.Org\ libVorbis\ I\ 20150105\ \(\342\233\204\342\233\204\342\233\204\342\233\204\)&quot;, encoder-version=(uint)0, audio-codec=(string)Vorbis, nominal-bitrate=(uint)80000, bitrate=(uint)80000, video-codec=(string)&quot;VP8\ video&quot;, file-size=(guint64)223340;' proxy-id='file:///nop/tears_of_steel.webm.223340.proxy.mov'/>
      <asset id='file:///nop/1sec_simpsons_trailer.mp4.232417.proxy.mov' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)6, duration=(guint64)1228020833;' metadatas='metadatas, container-format=(string)Matroska, audio-codec=(string)Opus, language-code=(string)en, encoder=(string)Lavf54.6.100, bitrate=(uint)64000, video-codec=(string)&quot;Motion\ JPEG&quot;, file-size=(guint64)4694708;' />
      <asset id='file:///nop/tears_of_steel.webm.223340.proxy.mov' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)6, duration=(guint64)2003000000;' metadatas='metadatas, container-format=(string)Matroska, language-code=(string)und, application-name=(string)Lavc56.60.100, encoder=(string)&quot;Xiph.Org\ libVorbis\ I\ 20150105\ \(\342\233\204\342\233\204\342\233\204\342\233\204\)&quot;, encoder-version=(uint)0, audio-codec=(string)Vorbis, nominal-bitrate=(uint)80000, bitrate=(uint)80000, video-codec=(string)&quot;VP8\ video&quot;, file-size=(guint64)223340;' />
    </ressources>
</project>
</ges>""")
        project_manager = app.project_manager
        medialib = MediaLibraryWidget(app)

        # Remove proxy
        with common.cloned_sample("1sec_simpsons_trailer.mp4", "tears_of_steel.webm"):
            def new_project_loaded_cb(*args, **kwargs):
                mainloop.quit()

            missing_uris = []

            def missing_uri_cb(project_manager, project, unused_error, asset):
                missing_uris.append(asset.props.id)
                return common.get_sample_uri(os.path.basename(asset.props.id))

            project_manager.connect("missing-uri", missing_uri_cb)
            project_manager.connect("new-project-loaded", new_project_loaded_cb)

            project_manager.load_project(proj_uri)
            mainloop.run()
            self.assertEqual(len(missing_uris), 1,
                             "missing_uri_cb should be called only once, got %s." % missing_uris)
            self.assertEqual(medialib._progressbar.get_fraction(), 1.0)
            mainloop.run()
            self.assertEqual(medialib.store.get_n_items(), 2,
                             "We should have one asset displayed in the MediaLibrary.")

            self.assertEqual(medialib.store[0].thumb_decorator.state, AssetThumbnail.PROXIED)
            self.assertEqual(medialib.store[1].thumb_decorator.state, AssetThumbnail.PROXIED)


class TestProjectSettings(common.TestCase):

    def test_audio(self):
        project = common.create_project()
        self.assertEqual(project.audiochannels, 2)
        self.assertEqual(project.audiorate, 96000)

        project.audiochannels = 6
        self.assertEqual(project.audiochannels, 6)
        project.audiorate = 44100
        self.assertEqual(project.audiorate, 44100)

    def test_video(self):
        project = common.create_project()
        self.assertEqual(project.videowidth, 1920)
        self.assertEqual(project.videoheight, 1080)
        self.assertEqual(project.videorate, Gst.Fraction(30, 1))

        project.videowidth = 3840
        self.assertEqual(project.videowidth, 3840)
        project.videoheight = 2160
        self.assertEqual(project.videoheight, 2160)
        project.videorate = Gst.Fraction(50, 7)
        self.assertEqual(project.videorate, Gst.Fraction(50, 7))

    def test_set_audio_prop(self):
        timeline = common.create_timeline_container()
        project = timeline.app.project_manager.current_project
        project.add_uris([common.get_sample_uri("mp3_sample.mp3")])

        audio_track = [t for t in project.ges_timeline.get_tracks() if isinstance(t, GES.AudioTrack)][0]
        mainloop = common.create_main_loop()

        def progress_cb(project, progress, estimated_time):
            if progress == 100:
                mainloop.quit()

        project.connect_after("asset-loading-progress", progress_cb)
        mainloop.run()

        expected = Gst.Caps("audio/x-raw,channels=(int)2,rate=(int)44100")
        ccaps = audio_track.props.restriction_caps
        self.assertTrue(ccaps.is_equal_fixed(expected), "%s != %s" % (ccaps, expected))

        project.audiochannels = 6

        expected = Gst.Caps("audio/x-raw,channels=(int)6,rate=(int)44100")
        ccaps = audio_track.props.restriction_caps
        self.assertTrue(ccaps.is_equal_fixed(expected), "%s != %s" % (ccaps, expected))

    def test_initialization(self):
        mainloop = common.create_main_loop()
        uris = collections.deque([
            common.get_sample_uri("flat_colour1_640x480.png"),
            common.get_sample_uri("tears_of_steel.webm"),
            common.get_sample_uri("1sec_simpsons_trailer.mp4")])

        def loaded_cb(project, timeline):
            project.add_uris([uris.popleft()])

        def progress_cb(project, progress, estimated_time):
            if progress == 100:
                if uris:
                    project.add_uris([uris.popleft()])
                else:
                    mainloop.quit()

        # Create a blank project and add some assets.
        project = common.create_project()
        self.assertTrue(project._has_default_video_settings)
        self.assertTrue(project._has_default_audio_settings)

        project.connect_after("loaded", loaded_cb)
        project.connect_after("asset-loading-progress", progress_cb)

        mainloop.run()

        assets = project.list_assets(GES.UriClip)
        self.assertEqual(3, len(assets), assets)

        self.assertFalse(project._has_default_video_settings)
        self.assertFalse(project._has_default_audio_settings)

        # The audio settings should match tears_of_steel.webm
        self.assertEqual(1, project.audiochannels)
        self.assertEqual(44100, project.audiorate)

        # The video settings should match tears_of_steel.webm
        self.assertEqual(960, project.videowidth)
        self.assertEqual(400, project.videoheight)
        self.assertEqual(Gst.Fraction(24, 1), project.videorate)

    def test_load(self):
        project = Project(uri="fake.xges", app=common.create_pitivi_mock())
        self.assertEqual(project.name, "fake")
        self.assertFalse(project._has_default_video_settings)
        self.assertFalse(project._has_default_audio_settings)

    def test_name(self):
        project = Project(common.create_pitivi_mock())

        project.uri = "file:///tmp/A%20B.xges"
        self.assertEqual(project.name, "A B")

        project.uri = "file:///tmp/%40%23%24%5E%26%60.xges"
        self.assertEqual(project.name, "@#$^&`")

    def test_scaled_proxy_size(self):
        app = common.create_pitivi_mock(default_scaled_proxy_width=123,
                                        default_scaled_proxy_height=456)
        manager = ProjectManager(app)
        project = manager.new_blank_project()
        self.assertFalse(project.has_scaled_proxy_size())
        self.assertEqual(project.scaled_proxy_width, 123)
        self.assertEqual(project.scaled_proxy_height, 456)

        with tempfile.NamedTemporaryFile() as temp_file:
            uri = Gst.filename_to_uri(temp_file.name)
            manager.save_project(uri=uri, backup=False)
            app2 = common.create_pitivi_mock(default_scaled_proxy_width=12,
                                             default_scaled_proxy_height=45)
            project2 = ProjectManager(app2).load_project(uri)
            self.assertFalse(project2.has_scaled_proxy_size())
            self.assertEqual(project2.scaled_proxy_width, 12)
            self.assertEqual(project2.scaled_proxy_height, 45)

        project.scaled_proxy_width = 123
        project.scaled_proxy_height = 456
        self.assertTrue(project.has_scaled_proxy_size())
        self.assertEqual(project.scaled_proxy_width, 123)
        self.assertEqual(project.scaled_proxy_height, 456)

        with tempfile.NamedTemporaryFile() as temp_file:
            manager.save_project(uri=uri, backup=False)
            app2 = common.create_pitivi_mock(default_scaled_proxy_width=1,
                                             default_scaled_proxy_height=4)
            project2 = ProjectManager(app2).load_project(uri)
            self.assertTrue(project2.has_scaled_proxy_size())
            self.assertEqual(project2.scaled_proxy_width, 123)
            self.assertEqual(project2.scaled_proxy_height, 456)


class TestExportSettings(common.TestCase):

    def test_master_attributes(self):
        self._check_master_attribute("muxer", dependant_attr="containersettings")
        self._check_master_attribute("vencoder", dependant_attr="vcodecsettings")
        self._check_master_attribute("aencoder", dependant_attr="acodecsettings")

    def _check_master_attribute(self, attr, dependant_attr):
        """Test changing the specified attr has effect on its dependent attr."""
        project = common.create_project()

        attr_value1 = "%s_value1" % attr
        attr_value2 = "%s_value2" % attr

        setattr(project, attr, attr_value1)
        setattr(project, dependant_attr, {})
        getattr(project, dependant_attr)["key1"] = "v1"

        setattr(project, attr, attr_value2)
        setattr(project, dependant_attr, {})
        getattr(project, dependant_attr)["key2"] = "v2"

        setattr(project, attr, attr_value1)
        self.assertTrue("key1" in getattr(project, dependant_attr))
        self.assertFalse("key2" in getattr(project, dependant_attr))
        self.assertEqual("v1", getattr(project, dependant_attr)["key1"])
        setattr(project, dependant_attr, {})

        setattr(project, attr, attr_value2)
        self.assertFalse("key1" in getattr(project, dependant_attr))
        self.assertTrue("key2" in getattr(project, dependant_attr))
        self.assertEqual("v2", getattr(project, dependant_attr)["key2"])
        setattr(project, dependant_attr, {})

        setattr(project, attr, attr_value1)
        self.assertFalse("key1" in getattr(project, dependant_attr))
        self.assertFalse("key2" in getattr(project, dependant_attr))

        setattr(project, attr, attr_value2)
        self.assertFalse("key1" in getattr(project, dependant_attr))
        self.assertFalse("key2" in getattr(project, dependant_attr))

    def test_set_rendering(self):
        """Checks the set_rendering method."""
        mainloop = common.create_main_loop()

        def loaded_cb(project, timeline):
            project.add_uris([common.get_sample_uri("tears_of_steel.webm")])

        def progress_cb(project, progress, estimated_time):
            if progress == 100:
                mainloop.quit()

        # Create a blank project and add some assets.
        project = common.create_project()

        project.connect_after("loaded", loaded_cb)
        project.connect_after("asset-loading-progress", progress_cb)

        mainloop.run()

        # The video settings should match tears_of_steel.webm
        self.assertEqual(project.videowidth, 960)
        self.assertEqual(project.videoheight, 400)

        project.render_scale = 3
        # Pretend we're rendering.
        project.set_rendering(True)
        self.assertEqual(project.videowidth, 28)
        self.assertEqual(project.videoheight, 12)

        # Pretend we're not rendering anymore.
        project.set_rendering(False)
        self.assertEqual(project.videowidth, 960)
        self.assertEqual(project.videoheight, 400)

    def test_set_safe_area_sizes(self):
        """Checks to ensure that the safe areas values are set correctly."""
        project = common.create_project()
        title_horizontal_factor = 0.8
        title_vertical_factor = 0.9
        action_horizontal_factor = 0.6
        action_vertical_factor = 0.7

        project.set_safe_areas_sizes(title_horizontal_factor,
                                     title_vertical_factor,
                                     action_horizontal_factor,
                                     action_vertical_factor)

        self.assertEqual(project.title_safe_area_horizontal, title_horizontal_factor)
        self.assertEqual(project.title_safe_area_vertical, title_vertical_factor)
        self.assertEqual(project.action_safe_area_horizontal, action_horizontal_factor)
        self.assertEqual(project.action_safe_area_vertical, action_vertical_factor)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2015, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the pitivi.timeline.timeline module."""
# pylint: disable=protected-access
from unittest import mock

from gi.repository import Gdk
from gi.repository import GES
from gi.repository import Gst
from gi.repository import Gtk

from pitivi.undo.timeline import TimelineObserver
from pitivi.undo.undo import UndoableActionLog
from pitivi.utils.timeline import UNSELECT
from pitivi.utils.ui import LAYER_HEIGHT
from pitivi.utils.ui import SEPARATOR_HEIGHT
from pitivi.utils.ui import URI_TARGET_ENTRY
from tests import common

THIN = LAYER_HEIGHT / 2
THICK = LAYER_HEIGHT


class TestLayers(common.TestCase):
    """Tests for the layers."""

    def test_dragging_layer(self):
        self.check_get_layer_at([THIN, THIN, THIN], 1, True,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])
        self.check_get_layer_at([THICK, THICK, THICK], 1, True,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])
        self.check_get_layer_at([THIN, THICK, THIN], 1, True,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])
        self.check_get_layer_at([THICK, THIN, THICK], 1, True,
                                [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2])

    def test_dragging_clip_from_layer(self):
        self.check_get_layer_at([THIN, THIN, THIN], 1, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])
        self.check_get_layer_at([THICK, THICK, THICK], 1, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])
        self.check_get_layer_at([THIN, THICK, THIN], 1, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])
        self.check_get_layer_at([THICK, THIN, THICK], 1, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2])

    def test_dragging_clip_from_outer_space(self):
        self.check_get_layer_at([THIN, THIN, THIN], None, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2])
        self.check_get_layer_at([THICK, THICK, THICK], None, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2])
        self.check_get_layer_at([THIN, THICK, THIN], None, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2])
        self.check_get_layer_at([THICK, THIN, THICK], None, False,
                                [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2])

    def check_get_layer_at(self, heights, preferred, past_middle_when_adjacent, expectations):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline

        # Allocate layers
        y = 0
        for priority, height in enumerate(heights):
            ges_layer = timeline.create_layer(priority=priority)
            rect = Gdk.Rectangle()
            rect.y = y
            rect.height = height
            ges_layer.ui.set_allocation(rect)
            y += height + SEPARATOR_HEIGHT

        ges_layers = timeline.ges_timeline.get_layers()
        if preferred is None:
            preferred_ges_layer = None
        else:
            preferred_ges_layer = ges_layers[preferred]
        # The heights of the layers.
        h = [ges_layer.ui.get_allocation().height for ges_layer in ges_layers]

        def assert_layer_at(ges_layer, y):
            result = timeline.get_layer_at(
                int(y),
                prefer_ges_layer=preferred_ges_layer,
                past_middle_when_adjacent=past_middle_when_adjacent)
            self.assertEqual(
                ges_layer,
                result[0],
                "Expected %d, got %d at %d" % (ges_layers.index(ges_layer), ges_layers.index(result[0]), y))

        # y on the top layer.
        assert_layer_at(ges_layers[expectations[0]], 0)
        assert_layer_at(ges_layers[expectations[1]], h[0] / 2 - 1)
        assert_layer_at(ges_layers[expectations[2]], h[0] / 2)
        assert_layer_at(ges_layers[expectations[3]], h[0] - 1)

        # y on the separator.
        assert_layer_at(ges_layers[expectations[4]], h[0])
        assert_layer_at(ges_layers[expectations[5]], h[0] + SEPARATOR_HEIGHT - 1)

        # y on the middle layer.
        assert_layer_at(ges_layers[expectations[6]], h[0] + SEPARATOR_HEIGHT)
        assert_layer_at(ges_layers[expectations[7]], h[0] + SEPARATOR_HEIGHT + h[1] / 2 - 1)
        assert_layer_at(ges_layers[expectations[8]], h[0] + SEPARATOR_HEIGHT + h[1] / 2)
        assert_layer_at(ges_layers[expectations[9]], h[0] + SEPARATOR_HEIGHT + h[1] - 1)

        # y on the separator.
        assert_layer_at(ges_layers[expectations[10]], h[0] + SEPARATOR_HEIGHT + h[1])
        assert_layer_at(ges_layers[expectations[11]], h[0] + SEPARATOR_HEIGHT + h[1] + SEPARATOR_HEIGHT - 1)

        # y on the bottom layer.
        assert_layer_at(ges_layers[expectations[12]], h[0] + SEPARATOR_HEIGHT + h[1] + SEPARATOR_HEIGHT)
        assert_layer_at(ges_layers[expectations[13]], h[0] + SEPARATOR_HEIGHT + h[1] + SEPARATOR_HEIGHT + h[2] / 2 - 1)
        assert_layer_at(ges_layers[expectations[14]], h[0] + SEPARATOR_HEIGHT + h[1] + SEPARATOR_HEIGHT + h[2] / 2)
        assert_layer_at(ges_layers[expectations[15]], h[0] + SEPARATOR_HEIGHT + h[1] + SEPARATOR_HEIGHT + h[2] - 1)

    def test_set_separators_prelight(self):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        timeline.__on_separators = [mock.Mock()]
        timeline._set_separators_prelight(False)
        self.assertEqual(len(timeline.__on_separators), 1,
                         "The separators must be forgotten only in drag_end()")

    def test_media_types(self):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline

        ges_layer_1 = timeline.ges_timeline.append_layer()
        ges_layer_2 = timeline.ges_timeline.append_layer()

        # Timeline should contain no media_types.
        self.assertEqual(timeline.media_types, GES.TrackType(0))

        # Timeline should now contain only audio media type.
        ges_clip_audio = self.add_clip(ges_layer_1, 10, clip_type=GES.TrackType.AUDIO)
        self.assertEqual(timeline.media_types, GES.TrackType.AUDIO)

        ges_layer_1.remove_clip(ges_clip_audio)
        ges_clip_video = self.add_clip(ges_layer_2, 20, clip_type=GES.TrackType.VIDEO)
        self.assertEqual(timeline.media_types, GES.TrackType.VIDEO)

        # Timeline should contain no media_types.
        ges_layer_2.remove_clip(ges_clip_video)
        self.assertEqual(timeline.media_types, GES.TrackType(0))

        # Timeline should contain both clips.
        ges_clip_audio = self.add_clip(ges_layer_1, 10, clip_type=GES.TrackType.AUDIO)
        ges_clip_video = self.add_clip(ges_layer_2, 20, clip_type=GES.TrackType.VIDEO)
        self.assertEqual(timeline.media_types,
                         GES.TrackType.VIDEO | GES.TrackType.AUDIO)

        # Timeline should contain no clips.
        ges_layer_1.remove_clip(ges_clip_audio)
        ges_layer_2.remove_clip(ges_clip_video)
        self.assertEqual(timeline.media_types, GES.TrackType(0))

    def test_create_layer(self):
        self.check_create_layer([0, 0, 0, 0], [3, 2, 1, 0])
        self.check_create_layer([0, 1, 1, 1], [0, 3, 2, 1])
        self.check_create_layer([0, 1, 1, 2], [0, 3, 1, 2])
        self.check_create_layer([0, 1, 0, 2], [1, 3, 0, 2])
        self.check_create_layer([0, 1, 2, 3], [0, 1, 2, 3])

    def check_create_layer(self, start_priorities, expected_priorities):
        timeline = common.create_timeline_container().timeline
        ges_layers = []
        for priority in start_priorities:
            ges_layer = timeline.create_layer(priority)
            self.assertEqual(ges_layer.props.priority, priority)
            ges_layers.append(ges_layer)
        self.check_priorities_and_positions(timeline, ges_layers, expected_priorities)

    def test_remove_layer(self):
        self.check_remove_layer([0, 0, 0])
        self.check_remove_layer([0, 0, 1])
        self.check_remove_layer([0, 1, 0])
        self.check_remove_layer([0, 2, 1])
        self.check_remove_layer([1, 0, 1])
        self.check_remove_layer([2, 2, 0])
        self.check_remove_layer([3, 2, 1])

    def check_remove_layer(self, removal_order):
        timeline = common.create_timeline_container().timeline

        # Add layers to remove them later.
        ges_layers = []
        # Pitivi doesn't support removing the last remaining layer,
        # that's why we create an extra layer.
        for priority in range(len(removal_order) + 1):
            ges_layer = timeline.create_layer(priority)
            ges_layers.append(ges_layer)

        # Remove layers one by one in the specified order.
        for priority in removal_order:
            ges_layer = ges_layers[priority]
            ges_layer.control_ui.delete_layer_action.activate(None)
            ges_layers.remove(ges_layer)
            self.check_priorities_and_positions(timeline, ges_layers, list(range(len(ges_layers))))

    def test_move_layer(self):
        self.check_move_layer(0, 0, [0, 1, 2, 3, 4])
        self.check_move_layer(0, 1, [1, 0, 2, 3, 4])
        self.check_move_layer(0, 4, [4, 0, 1, 2, 3])
        self.check_move_layer(2, 0, [1, 2, 0, 3, 4])
        self.check_move_layer(2, 3, [0, 1, 3, 2, 4])
        self.check_move_layer(4, 0, [1, 2, 3, 4, 0])
        self.check_move_layer(4, 3, [0, 1, 2, 4, 3])

    def check_move_layer(self, from_priority, to_priority, expected_priorities):
        timeline = common.create_timeline_container().timeline

        # Add layers to move them later.
        ges_layers = []
        for priority in range(len(expected_priorities)):
            ges_layer = timeline.create_layer(priority)
            ges_layers.append(ges_layer)

        timeline.move_layer(ges_layers[from_priority], to_priority)
        self.check_priorities_and_positions(timeline, ges_layers, expected_priorities)


class TestGrouping(common.TestCase):

    def __check_can_group_ungroup(self, timeline_container, can_group, can_ungroup):
        self.assertEqual(timeline_container.group_action.props.enabled, can_group)
        self.assertEqual(timeline_container.ungroup_action.props.enabled, can_ungroup)

    def test_can_group_ungroup(self):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        self.__check_can_group_ungroup(timeline_container, False, False)
        ges_clip, = self.add_clips_simple(timeline, 1)
        self.toggle_clip_selection(ges_clip, expect_selected=True)
        self.__check_can_group_ungroup(timeline_container, False, True)

        timeline_container.ungroup_action.emit("activate", None)
        self.__check_can_group_ungroup(timeline_container, False, False)

        layer, = timeline.ges_timeline.get_layers()
        ges_clip0, ges_clip1 = layer.get_clips()
        self.toggle_clip_selection(ges_clip0, expect_selected=True)
        self.__check_can_group_ungroup(timeline_container, False, False)

        # Press <ctrl> so selecting in ADD mode
        event = mock.Mock()
        event.keyval = Gdk.KEY_Control_L
        timeline_container.do_key_press_event(event)

        self.toggle_clip_selection(ges_clip1, expect_selected=True)
        self.__check_can_group_ungroup(timeline_container, True, False)

        timeline_container.group_action.emit("activate", None)
        self.__check_can_group_ungroup(timeline_container, False, True)

    def group_clips(self, timeline_container, clips):
        timeline = timeline_container.timeline
        timeline.app.settings.leftClickAlsoSeeks = False

        # Press <ctrl> so selecting in ADD mode
        event = mock.Mock()
        event.keyval = Gdk.KEY_Control_L
        timeline_container.do_key_press_event(event)
        timeline.get_clicked_layer_and_pos = mock.Mock()
        timeline.get_clicked_layer_and_pos.return_value = (None, None)

        # Select the 2 clips
        for clip in clips:
            self.assertIsNone(clip.get_parent())
            self.toggle_clip_selection(clip, expect_selected=True)

        timeline_container.group_action.emit("activate", None)

        for clip in clips:
            # Check that we created a new group
            self.assertTrue(isinstance(clip.get_parent(), GES.Group))
            # The newly created group has been selected
            for selected_clip in timeline.selection:
                self.assertEqual(clip.get_toplevel_parent(), selected_clip.get_toplevel_parent())

            self.assertEqual(clips[0].get_parent(), clip.get_parent())
            self.assertTrue(bool(clip.ui.get_state_flags() & Gtk.StateFlags.SELECTED))
            self.assertTrue(clip.selected)

        group = clips[0].get_parent()
        self.assertEqual(len(group.get_children(False)), len(clips))

    def test_group(self):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clips = self.add_clips_simple(timeline, 2)
        self.group_clips(timeline_container, clips)

    def test_group_selection(self):
        num_clips = 2
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clips = self.add_clips_simple(timeline, num_clips)
        self.group_clips(timeline_container, clips)
        layer = timeline.ges_timeline.get_layers()[0]
        clips = layer.get_clips()
        self.assertEqual(len(clips), num_clips)

        # Deselect one grouped clip clips
        self.toggle_clip_selection(clips[0], expect_selected=False)

        # Make sure all the clips have been deselected
        for clip in clips:
            self.assertFalse(bool(clip.ui.get_state_flags() & Gtk.StateFlags.SELECTED))
            self.assertFalse(clip.selected)

    def test_group_ungroup(self):
        num_clips = 2
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clips = self.add_clips_simple(timeline, num_clips)
        self.group_clips(timeline_container, clips)

        self.assertEqual(len(timeline.selection.selected), num_clips)

        timeline_container.ungroup_action.emit("activate", None)
        layer = timeline.ges_timeline.get_layers()[0]
        clips = layer.get_clips()
        self.assertEqual(len(clips), num_clips)

        for clip in clips:
            self.assertIsNone(clip.get_parent())

    def test_group_splitted_clip_and_select_group(self):

        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clips = self.add_clips_simple(timeline, 1)
        clips[0].props.duration = timeline.ges_timeline.get_frame_time(4)
        position = timeline.ges_timeline.get_frame_time(2)
        self.toggle_clip_selection(clips[0], expect_selected=True)

        timeline.ges_timeline.get_asset().pipeline.get_position = mock.Mock(return_value=position)
        layer = timeline.ges_timeline.get_layers()[0]

        # Split
        timeline_container.split_action.emit("activate", None)
        clips = layer.get_clips()
        self.assertEqual(len(clips), 2)

        # Only the first clip is selected so select the
        # second one
        self.assertTrue(clips[0].selected)
        self.assertFalse(clips[1].selected)

        event = mock.Mock()
        event.keyval = Gdk.KEY_Control_L
        timeline_container.do_key_press_event(event)
        timeline.get_clicked_layer_and_pos = mock.Mock()
        timeline.get_clicked_layer_and_pos.return_value = (None, None)
        self.toggle_clip_selection(clips[1], expect_selected=True)
        timeline_container.do_key_release_event(event)

        for clip in clips:
            self.assertTrue(clip.selected)

        # Group the two parts
        timeline_container.group_action.emit("activate", None)

        self.toggle_clip_selection(clips[1], expect_selected=True)

    def test_ungroup_clip(self):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        ges_clip, = self.add_clips_simple(timeline, 1)

        self.toggle_clip_selection(ges_clip, expect_selected=True)

        timeline_container.ungroup_action.emit("activate", None)
        layer = timeline.ges_timeline.get_layers()[0]
        ges_clip0, ges_clip1 = layer.get_clips()

        self.assertEqual(ges_clip0.props.start, ges_clip1.props.start)
        self.assertEqual(ges_clip0.props.duration, ges_clip1.props.duration)

        track_elem_0, = ges_clip0.get_children(recursive=False)
        track_elem_1, = ges_clip1.get_children(recursive=False)

        if track_elem_0.get_track_type() == GES.TrackType.AUDIO:
            aclip = ges_clip0.ui
            atrackelem = track_elem_0.ui
            vclip = ges_clip1.ui
            vtrackelem = track_elem_1.ui
        else:
            aclip = ges_clip1.ui
            atrackelem = track_elem_1.ui

            vclip = ges_clip0.ui
            vtrackelem = track_elem_0.ui

        self.assertEqual(aclip.audio_widget, atrackelem)
        self.assertEqual(vclip.video_widget, vtrackelem)

    def test_dragging_group_on_separator(self):
        # Create two clips on different layers and group them.
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clip1, = self.add_clips_simple(timeline, 1)
        layer1 = clip1.get_layer()

        # Add another clip on a new layer.
        clip2, = self.add_clips_simple(timeline, 1)
        self.assertEqual(len(timeline.ges_timeline.get_layers()), 2)

        self.group_clips(timeline_container, [clip1, clip2])

        # Click the first clip in the group.
        with mock.patch.object(Gtk, 'get_event_widget') as get_event_widget:
            event = mock.Mock()
            event.x = 100
            event.get_button.return_value = True, 1
            get_event_widget.return_value = clip1.ui
            timeline._button_press_event_cb(None, event)
            self.assertIsNotNone(timeline.dragging_element)

            # Move it to the right, on the separator below.
            event = mock.Mock()
            event.x = 101
            event.get_state.return_value = Gdk.ModifierType.BUTTON1_MASK
            with mock.patch.object(clip1.ui, "translate_coordinates") as translate_coordinates:
                translate_coordinates.return_value = (40, 0)
                with mock.patch.object(timeline, "get_layer_at") as get_layer_at:
                    get_layer_at.return_value = layer1, timeline._separators[1]
                    timeline._motion_notify_event_cb(None, event)
            self.assertTrue(timeline.got_dragged)

        # Release the mouse button.
        event = mock.Mock()
        event.get_button.return_value = True, 1
        timeline._button_release_event_cb(None, event)
        self.assertEqual(len(timeline.ges_timeline.get_layers()), 2,
                         "No new layer should have been created")


class TestCopyPaste(common.TestCase):

    def copy_clips(self, num_clips):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline

        clips = self.add_clips_simple(timeline, num_clips)

        # Press <ctrl> so selecting in ADD mode
        event = mock.Mock()
        event.keyval = Gdk.KEY_Control_L
        timeline_container.do_key_press_event(event)
        timeline.get_clicked_layer_and_pos = mock.Mock()
        timeline.get_clicked_layer_and_pos.return_value = (None, None)

        # Select the 2 clips
        for clip in clips:
            self.toggle_clip_selection(clip, expect_selected=True)

        self.assertTrue(timeline_container.copy_action.props.enabled)
        self.assertFalse(timeline_container.paste_action.props.enabled)
        timeline_container.copy_action.emit("activate", None)
        self.assertTrue(timeline_container.paste_action.props.enabled)

        return timeline_container

    def test_copy_paste(self):
        timeline_container = self.copy_clips(2)
        timeline = timeline_container.timeline
        layer = timeline.ges_timeline.get_layers()[0]
        project = timeline.ges_timeline.get_asset()

        clips = layer.get_clips()
        self.assertEqual(len(clips), 2)

        # Pasting clips for the first time.
        position = 20
        project.pipeline.get_position = mock.Mock(return_value=position)
        timeline_container.paste_action.emit("activate", None)

        n_clips = layer.get_clips()
        self.assertEqual(len(n_clips), 4)

        copied_clips = [clip for clip in n_clips if clip not in clips]
        self.assertEqual(len(copied_clips), 2)
        self.assertEqual(copied_clips[0].props.start, position)
        self.assertEqual(copied_clips[1].props.start, position + 10)

        # Pasting same clips second time.
        position = 40
        project.pipeline.get_position = mock.Mock(return_value=position)
        timeline_container.paste_action.emit("activate", None)

        n_clips = layer.get_clips()
        self.assertEqual(len(n_clips), 6)

        copied_clips = [clip for clip in n_clips if clip not in clips]
        self.assertEqual(len(copied_clips), 4)
        self.assertEqual(copied_clips[2].props.start, position)
        self.assertEqual(copied_clips[3].props.start, position + 10)

    def test_paste_not_possible(self):
        timeline_container = self.copy_clips(1)
        timeline = timeline_container.timeline
        layer = timeline.ges_timeline.get_layers()[0]
        project = timeline.ges_timeline.get_asset()
        self.assertEqual(len(layer.get_clips()), 1)

        position = 0
        project.pipeline.get_position = mock.Mock(return_value=position)
        timeline_container.paste_action.emit("activate", None)
        self.assertEqual(len(layer.get_clips()), 1)

    def test_paste_selection(self):
        timeline_container = self.copy_clips(1)
        timeline = timeline_container.timeline
        layer = timeline.ges_timeline.get_layers()[0]
        project = timeline.ges_timeline.get_asset()

        # Paste a single clip.
        clips = layer.get_clips()
        self.assertEqual(len(clips), 1)
        position = 10
        project.pipeline.get_position = mock.Mock(return_value=position)
        timeline_container.paste_action.emit("activate", None)

        clips = layer.get_clips()
        self.assertListEqual([bool(clip.selected) for clip in clips], [False, True])

        # The copy_clips() above simulates CTRL button press, so this will add 1st clip to the selection.
        self.toggle_clip_selection(clips[0], True)
        timeline_container.copy_action.emit("activate", None)

        # Paste two clips.
        position = 20
        project.pipeline.get_position = mock.Mock(return_value=position)
        timeline_container.paste_action.emit("activate", None)

        clips = layer.get_clips()
        self.assertListEqual([bool(clip.selected) for clip in clips], [False, False, True, True])


class TestEditing(common.TestCase):

    def test_trimming_on_layer_separator(self):
        # Create a clip
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clip, = self.add_clips_simple(timeline, 1)
        layer = clip.get_layer()

        # Click the right trim handle of the clip.
        with mock.patch.object(Gtk, 'get_event_widget') as get_event_widget:
            event = mock.Mock()
            event.x = 100
            event.get_button.return_value = True, 1
            get_event_widget.return_value = clip.ui.right_handle
            timeline._button_press_event_cb(None, event)
            self.assertIsNotNone(timeline.dragging_element)

            # Drag it to the left, on the separator below.
            event = mock.Mock()
            event.x = 99
            event.get_state.return_value = Gdk.ModifierType.BUTTON1_MASK
            with mock.patch.object(clip.ui.right_handle, "translate_coordinates") as translate_coordinates:
                translate_coordinates.return_value = (0, 0)
                with mock.patch.object(timeline, "get_layer_at") as get_layer_at:
                    get_layer_at.return_value = layer, timeline._separators[1]
                    timeline._motion_notify_event_cb(None, event)
            self.assertTrue(timeline.got_dragged)

        # Release the mouse button.
        event = mock.Mock()
        event.get_button.return_value = True, 1
        timeline._button_release_event_cb(None, event)
        self.assertEqual(len(timeline.ges_timeline.get_layers()), 1,
                         "No new layer should have been created")


class TestShiftSelection(common.TestCase):

    def __reset_clips_selection(self, timeline):
        """Unselects all clips in the timeline."""
        layers = timeline.ges_timeline.get_layers()
        for layer in layers:
            clips = layer.get_clips()
            timeline.selection.set_selection(clips, UNSELECT)
            timeline.set_selection_meta_info(layer, 0, UNSELECT)

    def __check_selected(self, selected_clips, not_selected_clips):
        for clip in selected_clips:
            self.assertEqual(clip.selected._selected, True)
        for clip in not_selected_clips:
            self.assertEqual(clip.selected._selected, False)

    def __check_simple(self, left_click_also_seeks):
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        timeline.app.settings.leftClickAlsoSeeks = left_click_also_seeks
        ges_layer = timeline.ges_timeline.append_layer()
        asset = GES.UriClipAsset.request_sync(
            common.get_sample_uri("1sec_simpsons_trailer.mp4"))
        ges_clip1 = ges_layer.add_asset(asset, 0 * Gst.SECOND, 0,
                                        1 * Gst.SECOND, GES.TrackType.UNKNOWN)
        ges_clip2 = ges_layer.add_asset(asset, 1 * Gst.SECOND, 0,
                                        1 * Gst.SECOND, GES.TrackType.UNKNOWN)

        event = mock.Mock()
        event.get_button.return_value = (True, 1)
        timeline._seek = mock.Mock()
        timeline._seek.return_value = True
        timeline.get_clicked_layer_and_pos = mock.Mock()

        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = timeline

            # Simulate click on first and shift+click on second clip.
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 0.5 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_parent().shift_mask = True
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 1.5 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip1, ges_clip2], [])

    def test_simple(self):
        self.__check_simple(left_click_also_seeks=False)
        self.__check_simple(left_click_also_seeks=True)

    def __check_shift_selection_single_layer(self, left_click_also_seeks):
        """Checks group clips selection across a single layer."""
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        timeline.app.settings.leftClickAlsoSeeks = left_click_also_seeks
        ges_layer = timeline.ges_timeline.append_layer()
        ges_clip1 = self.add_clip(ges_layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip2 = self.add_clip(ges_layer, 15 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip3 = self.add_clip(ges_layer, 25 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip4 = self.add_clip(ges_layer, 35 * Gst.SECOND, duration=2 * Gst.SECOND)

        event = mock.Mock()
        event.get_button.return_value = (True, 1)
        timeline.get_parent().shift_mask = True
        timeline._seek = mock.Mock()
        timeline._seek.return_value = True
        timeline.get_clicked_layer_and_pos = mock.Mock()

        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = timeline

            # Simulate shift+click before first and on second clip.
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 1 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 17 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip1, ges_clip2], [ges_clip3, ges_clip4])
            self.__reset_clips_selection(timeline)

            # Simiulate shift+click before first and after fourth clip.
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 1 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 39 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip1, ges_clip2, ges_clip3, ges_clip4], [])
            self.__reset_clips_selection(timeline)

            # Simiulate shift+click on first, after fourth and before third clip.
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 6 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 40 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 23 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip1, ges_clip2], [ges_clip3, ges_clip4])
            self.__reset_clips_selection(timeline)

            # Simulate shift+click twice on the same clip.
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 6 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer, 6.5 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip1], [ges_clip2, ges_clip3, ges_clip4])

    def test_shift_selection_single_layer(self):
        self.__check_shift_selection_single_layer(left_click_also_seeks=False)
        self.__check_shift_selection_single_layer(left_click_also_seeks=True)

    def __check_shift_selection_multiple_layers(self, left_click_also_seeks):
        """Checks group clips selection across multiple layers."""
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        timeline.app.settings.leftClickAlsoSeeks = left_click_also_seeks
        ges_layer1 = timeline.ges_timeline.append_layer()
        ges_clip11 = self.add_clip(ges_layer1, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip12 = self.add_clip(ges_layer1, 15 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip13 = self.add_clip(ges_layer1, 25 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_layer2 = timeline.ges_timeline.append_layer()
        ges_clip21 = self.add_clip(ges_layer2, 0 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip22 = self.add_clip(ges_layer2, 6 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip23 = self.add_clip(ges_layer2, 21 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_layer3 = timeline.ges_timeline.append_layer()
        ges_clip31 = self.add_clip(ges_layer3, 3 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip32 = self.add_clip(ges_layer3, 10 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip33 = self.add_clip(ges_layer3, 18 * Gst.SECOND, duration=2 * Gst.SECOND)

        event = mock.Mock()
        event.get_button.return_value = (True, 1)
        timeline.get_parent().shift_mask = True
        timeline._seek = mock.Mock()
        timeline._seek.return_value = True
        timeline.get_clicked_layer_and_pos = mock.Mock()

        with mock.patch.object(Gtk, "get_event_widget") as get_event_widget:
            get_event_widget.return_value = timeline

            timeline.get_clicked_layer_and_pos.return_value = (ges_layer2, 3 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer1, 9 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip11, ges_clip22],
                                  [ges_clip12, ges_clip13, ges_clip21, ges_clip23, ges_clip31, ges_clip32, ges_clip33])
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer3, 12 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip22, ges_clip31, ges_clip32],
                                  [ges_clip11, ges_clip12, ges_clip13, ges_clip21, ges_clip23, ges_clip33])
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer1, 22 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip11, ges_clip12, ges_clip22, ges_clip23],
                                  [ges_clip13, ges_clip21, ges_clip31, ges_clip32, ges_clip33])
            self.__reset_clips_selection(timeline)

            timeline.get_clicked_layer_and_pos.return_value = (ges_layer1, 3 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer2, 26 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip11, ges_clip12, ges_clip13, ges_clip22, ges_clip23],
                                  [ges_clip21, ges_clip31, ges_clip32, ges_clip33])
            timeline.get_clicked_layer_and_pos.return_value = (ges_layer3, 30 * Gst.SECOND)
            timeline._button_release_event_cb(None, event)
            self.__check_selected([ges_clip11, ges_clip12, ges_clip13, ges_clip22, ges_clip23,
                                   ges_clip31, ges_clip32, ges_clip33],
                                  [ges_clip21])

    def test_shift_selection_multiple_layers(self):
        self.__check_shift_selection_multiple_layers(left_click_also_seeks=False)
        self.__check_shift_selection_multiple_layers(left_click_also_seeks=True)


class TestTimelineContainer(common.TestCase):
    """Tests for the TimelineContainer class."""

    def test_update_clips_asset(self):
        timeline_container = common.create_timeline_container()
        mainloop = common.create_main_loop()
        mainloop.run(until_empty=True)
        ges_timeline = timeline_container.ges_timeline
        layer, = ges_timeline.get_layers()
        title_clip = GES.TitleClip()
        title_clip.props.duration = 100
        layer.add_clip(title_clip)
        self.assertListEqual(list(timeline_container.timeline.clips()), [title_clip])

        # Check the title clips are ignored.
        timeline_container.update_clips_asset(mock.Mock())


class TestClipsEdges(common.TestCase):

    def test_clips_edges(self):
        """Test function for function clips_edges."""
        timeline_container = common.create_timeline_container()
        timeline = timeline_container.timeline
        clips = self.add_clips_simple(timeline, 5)
        timeline.ges_timeline.layers[0].remove_clip(clips[0])
        self.add_clip(timeline.ges_timeline.layers[0], 5, 0, 10)

        self.assertEqual(timeline_container.first_clip_edge(after=0), 5)
        self.assertEqual(timeline_container.first_clip_edge(after=9), 10)
        self.assertEqual(timeline_container.first_clip_edge(after=10), 15)
        self.assertEqual(timeline_container.first_clip_edge(after=48), 50)
        self.assertEqual(timeline_container.first_clip_edge(after=49), 50)

        self.assertEqual(timeline_container.first_clip_edge(before=0), None)
        self.assertEqual(timeline_container.first_clip_edge(before=1), 0)
        self.assertEqual(timeline_container.first_clip_edge(before=9), 5)
        self.assertEqual(timeline_container.first_clip_edge(before=10), 5)
        self.assertEqual(timeline_container.first_clip_edge(before=11), 10)
        self.assertEqual(timeline_container.first_clip_edge(before=20), 15)


class TestDragFromOutside(common.TestCase):

    def setUp(self):
        super().setUp()

        timeline_container = common.create_timeline_container()
        self.ges_timeline = timeline_container.timeline.ges_timeline

        timeline_container.app.action_log = UndoableActionLog()
        self.timeline_observer = TimelineObserver(timeline_container.ges_timeline, timeline_container.app.action_log)

    def check_drag_assets_to_timeline(self, timeline_ui, assets):
        # Events emitted while dragging assets over a clip in the timeline:
        # motion, receive, motion.
        with mock.patch.object(Gdk, "drag_status") as _drag_status_mock:
            with mock.patch.object(Gtk, "drag_finish") as _drag_finish_mock:
                target = mock.Mock()
                target.name.return_value = URI_TARGET_ENTRY.target
                timeline_ui.drag_dest_find_target = mock.Mock(return_value=target)
                timeline_ui.drag_get_data = mock.Mock()
                timeline_ui._drag_motion_cb(None, None, 0, 0, 0)
                self.assertTrue(timeline_ui.drag_get_data.called)

                self.assertFalse(timeline_ui.drop_data_ready)
                selection_data = mock.Mock()
                selection_data.get_data_type = mock.Mock(return_value=target)
                selection_data.get_uris.return_value = [asset.props.id for asset in assets]
                self.assertIsNone(timeline_ui.drop_data)
                self.assertFalse(timeline_ui.drop_data_ready)
                timeline_ui._drag_data_received_cb(None, None, 0, 0, selection_data, None, 0)
                self.assertEqual(timeline_ui.drop_data, [asset.props.id for asset in assets])
                self.assertTrue(timeline_ui.drop_data_ready)

                timeline_ui.drag_get_data.reset_mock()
                self.assertIsNone(timeline_ui.dragging_element)
                self.assertFalse(timeline_ui.dropping_clips)

                # Drag on the first layer.
                def translate_coordinates_func(widget, x, y):
                    return x, y
                timeline_ui.translate_coordinates = translate_coordinates_func
                timeline_ui._drag_motion_cb(timeline_ui, None, 0, SEPARATOR_HEIGHT, 0)
                self.assertFalse(timeline_ui.drag_get_data.called)
                self.assertIsNone(timeline_ui.dragging_element)
                self.assertFalse(timeline_ui.dropping_clips)

    def test_adding_overlap_clip(self):
        """Checks asset drag&drop on top of an existing clip."""
        asset = GES.UriClipAsset.request_sync(
            common.get_sample_uri("tears_of_steel.webm"))

        layer, = self.ges_timeline.get_layers()
        layer.add_asset(asset, 0, 0, 10, GES.TrackType.UNKNOWN)
        clips = layer.get_clips()

        self.check_drag_assets_to_timeline(self.ges_timeline.ui, [asset])
        self.assertEqual(layer.get_clips(), clips)

    def test_dragging_multiple_clips_over_timeline(self):
        """Checks drag&drop two assets when only the first one can be placed."""
        asset = GES.UriClipAsset.request_sync(
            common.get_sample_uri("tears_of_steel.webm"))

        layer, = self.ges_timeline.get_layers()
        start = asset.get_duration()
        layer.add_asset(asset, start, 0, 10, GES.TrackType.UNKNOWN)
        clips = layer.get_clips()

        # Use same asset to mimic dragging multiple assets
        self.check_drag_assets_to_timeline(self.ges_timeline.ui, [asset, asset])
        self.assertEqual(layer.get_clips(), clips)


class TestKeyboardShiftClips(common.TestCase):

    def check_frame_shift_clips(self, *ges_clips):
        """Checks that clips shifted forwards then backwards work properly."""
        clip_original_starts = [clip.start for clip in ges_clips]
        delta = self.timeline.get_frame_time(1)

        event = mock.Mock()
        event.keyval = Gdk.KEY_Control_L
        self.timeline_container.do_key_press_event(event)
        for clip in ges_clips:
            self.toggle_clip_selection(clip, expect_selected=True)
        self.timeline_container.do_key_release_event(event)

        self.timeline_container.shift_forward_action.emit("activate", None)
        self.assertListEqual([clip.start - delta for clip in ges_clips], clip_original_starts)

        self.timeline_container.shift_backward_action.emit("activate", None)
        self.assertListEqual([clip.start for clip in ges_clips], clip_original_starts)

    @common.setup_project(["tears_of_steel.webm"])
    def test_clip_shift(self):
        """Checks that shift methods change position of a single clip by one frame."""
        ges_clip1 = self.add_clip(self.layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        self.check_frame_shift_clips(ges_clip1)

    @common.setup_project(["tears_of_steel.webm"])
    def test_shift_disjoint_clips(self):
        """Checks that disjoint clips are able to be shifted."""
        ges_clip1 = self.add_clip(self.layer, 5 * Gst.SECOND, duration=1 * Gst.SECOND)
        ges_clip2 = self.add_clip(self.layer, 9 * Gst.SECOND, duration=1 * Gst.SECOND)
        self.check_frame_shift_clips(ges_clip1, ges_clip2)

    @common.setup_project(["tears_of_steel.webm"])
    def test_shift_adjacent_clips(self):
        """Checks that adjacent clips are able to be shifted as well."""
        ges_clip1 = self.add_clip(self.layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip2 = self.add_clip(self.layer, 7 * Gst.SECOND, duration=2 * Gst.SECOND)
        self.check_frame_shift_clips(ges_clip1, ges_clip2)

    @common.setup_project(["tears_of_steel.webm"])
    def test_triple_overlap_causes_rollback(self):
        """Checks that rollback works properly in the event of triple overlap."""
        ges_clip1 = self.add_clip(self.layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip2 = self.add_clip(self.layer, 10 * Gst.SECOND, duration=2 * Gst.SECOND)
        self.add_clip(self.layer, start=11 * Gst.SECOND, duration=2 * Gst.SECOND)
        self.add_clip(self.layer, start=12 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip3 = self.add_clip(self.layer, 13 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip4 = self.add_clip(self.layer, 15 * Gst.SECOND, duration=2 * Gst.SECOND)

        event = mock.Mock()
        event.keyval = Gdk.KEY_Control_L
        self.timeline_container.do_key_press_event(event)
        self.toggle_clip_selection(ges_clip1, expect_selected=True)
        self.toggle_clip_selection(ges_clip2, expect_selected=True)
        self.toggle_clip_selection(ges_clip3, expect_selected=True)
        self.toggle_clip_selection(ges_clip4, expect_selected=True)

        self.timeline_container.do_key_release_event(event)

        self.timeline_container.shift_forward_action.emit("activate", None)

        self.assertEqual(5 * Gst.SECOND, ges_clip1.start)
        self.assertEqual(10 * Gst.SECOND, ges_clip2.start)
        self.assertEqual(13 * Gst.SECOND, ges_clip3.start)
        self.assertEqual(15 * Gst.SECOND, ges_clip4.start)

        self.timeline_container.shift_backward_action.emit("activate", None)

        self.assertEqual(5 * Gst.SECOND, ges_clip1.start)
        self.assertEqual(10 * Gst.SECOND, ges_clip2.start)
        self.assertEqual(13 * Gst.SECOND, ges_clip3.start)
        self.assertEqual(15 * Gst.SECOND, ges_clip4.start)


class TestSnapClips(common.TestCase):

    @common.setup_timeline
    def test_snap_clip_single_layer_single_clip(self):
        """Test whether a single clip is able to snap right, and left to an adjacent clip."""
        ges_clip1 = self.add_clip(self.layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip2 = self.add_clip(self.layer, 9 * Gst.SECOND, duration=2 * Gst.SECOND)

        self.toggle_clip_selection(ges_clip1, expect_selected=True)

        self.timeline_container.snap_clips_forward_action.emit("activate", None)
        self.assertEqual(ges_clip1.start, ges_clip2.start - ges_clip1.duration)

        self.timeline_container.snap_clips_backward_action.emit("activate", None)
        self.assertEqual(ges_clip1.start, 0)

    @common.setup_timeline
    def test_snap_single_layer_multiple_clips_adjacent(self):
        """Tests whether a single clip can snap to multiple adjacent clips."""
        ges_clip1 = self.add_clip(self.layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip2 = self.add_clip(self.layer, 7 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip3 = self.add_clip(self.layer, 11 * Gst.SECOND, duration=2 * Gst.SECOND)

        event = mock.Mock()
        event.keyval = Gdk.KEY_Control_L
        self.timeline_container.do_key_press_event(event)
        self.toggle_clip_selection(ges_clip1, expect_selected=True)
        self.toggle_clip_selection(ges_clip2, expect_selected=True)

        self.timeline_container.snap_clips_forward_action.emit("activate", None)
        self.assertEqual(ges_clip2.start, ges_clip3.start - ges_clip2.duration)
        self.assertEqual(ges_clip1.start, ges_clip2.start - ges_clip1.duration)

        self.timeline_container.snap_clips_backward_action.emit("activate", None)
        self.assertEqual(ges_clip1.start, 0)
        self.assertEqual(ges_clip2.start, ges_clip1.start + ges_clip1.duration)

    @common.setup_timeline
    def test_snap_multiple_layers_not_affected_by_other_layer(self):
        """Tests whether a clip snap is affected by a clip in another layer."""
        layer2 = self.timeline.append_layer()
        self.add_clip(self.layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        self.add_clip(self.layer, 7 * Gst.SECOND, duration=2 * Gst.SECOND)
        self.add_clip(self.layer, 11 * Gst.SECOND, duration=2 * Gst.SECOND)
        clip = self.add_clip(layer2, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        end_clip = self.add_clip(layer2, 30 * Gst.SECOND, duration=2 * Gst.SECOND)

        self.toggle_clip_selection(clip, expect_selected=True)

        self.timeline_container.snap_clips_forward_action.emit("activate", None)
        self.assertEqual(clip.start, end_clip.start - clip.duration)

        self.timeline_container.snap_clips_backward_action.emit("activate", None)
        self.assertEqual(clip.start, 0)

    @common.setup_timeline
    def test_single_clip_snap_right_does_nothing(self):
        """Tests whether the last clip snapped forward remains in place."""
        ges_clip = self.add_clip(self.layer, 5 * Gst.SECOND, duration=2 * Gst.SECOND)
        clip_start = ges_clip.start

        self.toggle_clip_selection(ges_clip, expect_selected=True)

        self.timeline_container.snap_clips_forward_action.emit("activate", None)
        self.assertEqual(ges_clip.start, clip_start)

    @common.setup_timeline
    def test_clip_snaps_to_timeline_duration(self):
        """Tests whether a clip snaps to the timeline duration."""
        layer2 = self.timeline.append_layer()
        self.add_clip(self.layer, 30 * Gst.SECOND, duration=2 * Gst.SECOND)
        ges_clip = self.add_clip(layer2, 5 * Gst.SECOND, duration=2 * Gst.SECOND)

        self.toggle_clip_selection(ges_clip, expect_selected=True)

        self.timeline_container.snap_clips_forward_action.emit("activate", None)
        self.assertEqual(ges_clip.start, self.timeline.get_duration() - ges_clip.duration)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2019, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Tests for the mainwindow module."""
# pylint: disable=no-self-use,protected-access
from unittest import mock

from pitivi.application import Pitivi
from pitivi.mainwindow import MainWindow
from tests import common


class MainWindowTest(common.TestCase):
    """Tests for the MainWindow class."""

    def test_create(self):
        """Exercise creating the main window."""
        app = Pitivi()
        app._setup()
        app.create_main_window()

    def test_create_small(self):
        """Exercise creating the main window when the screen is small."""
        with mock.patch.object(MainWindow, "_small_screen") as small_screen:
            small_screen.return_value = True
            with mock.patch.object(MainWindow, "get_preferred_size") as get:
                get.return_value = mock.Mock(), None
                app = Pitivi()
                app._setup()
                app.create_main_window()



# -*- coding: utf-8 -*-
# Pitivi Developer Console
# Copyright (c) 2017-2018, Fabian Orccon <cfoch.fabian@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""The developer console widget."""
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from plugins.console.consolebuffer import ConsoleBuffer


class ConsoleView(Gtk.TextView):
    """A TextView which removes tags from pasted text."""

    def do_paste_clipboard(self):
        # pylint: disable=arguments-differ
        buf = self.get_buffer()
        insert = buf.get_insert()
        paste_mark = buf.create_mark("paste-mark", buf.get_iter_at_mark(insert),
                                     left_gravity=True)

        Gtk.TextView.do_paste_clipboard(self)

        start = buf.get_iter_at_mark(paste_mark)
        end = buf.get_iter_at_mark(insert)

        buf.remove_all_tags(start, end)
        buf.delete_mark(paste_mark)


class ConsoleWidget(Gtk.ScrolledWindow):
    """An emulated Python console.

    The console can be used to access an app, window, or anything through the
    provided namespace. It works redirecting stdout and stderr to a
    GtkTextBuffer. This class is (and should be) independent of the application
    it is integrated with.
    """

    __gsignals__ = {
        "eof": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, namespace, welcome_message=""):
        Gtk.ScrolledWindow.__init__(self)
        self._view = ConsoleView()
        buf = ConsoleBuffer(namespace, welcome_message)
        self._view.set_buffer(buf)
        self._view.set_editable(True)
        self.add(self._view)

        self._view.connect("key-press-event", self.__key_press_event_cb)
        buf.connect("mark-set", self.__mark_set_cb)
        buf.connect("insert-text", self.__insert_text_cb)

        # Font color and style.
        self._provider = Gtk.CssProvider()
        self._css_values = {
            "textview": {
                "font-family": None,
                "font-size": None,
                "font-style": None,
                "font-variant": None,
                "font-weight": None
            },
            "textview > *": {
                "color": None
            }
        }

    def scroll_to_end(self):
        """Scrolls the view to the end."""
        end_iter = self._view.get_buffer().get_end_iter()
        self._view.scroll_to_iter(end_iter, within_margin=0.0, use_align=False,
                                  xalign=0, yalign=0)
        return False

    def set_font(self, font_desc):
        """Sets the font.

        Args:
            font_desc (str): a PangoFontDescription as a string.
        """
        pango_font_desc = Pango.FontDescription.from_string(font_desc)
        self._css_values["textview"]["font-family"] = pango_font_desc.get_family()
        self._css_values["textview"]["font-size"] = "%dpt" % int(pango_font_desc.get_size() / Pango.SCALE)
        self._css_values["textview"]["font-style"] = pango_font_desc.get_style().value_nick
        self._css_values["textview"]["font-variant"] = pango_font_desc.get_variant().value_nick
        self._css_values["textview"]["font-weight"] = int(pango_font_desc.get_weight())
        self._apply_css()

    def set_color(self, color):
        """Sets the color.

        Args:
            color (Gdk.RGBA): a color.
        """
        self._css_values["textview > *"]["color"] = color.to_string()
        self._apply_css()

    def _apply_css(self):
        css = ""
        for css_klass, props in self._css_values.items():
            css += "%s {" % css_klass
            for prop, value in props.items():
                if value is not None:
                    css += "%s: %s;" % (prop, value)
            css += "} "
        css = css.encode("UTF-8")
        self._provider.load_from_data(css)
        Gtk.StyleContext.add_provider(self._view.get_style_context(),
                                      self._provider,
                                      Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def set_stdout_color(self, color):
        """Sets the color of the stdout text."""
        self._view.get_buffer().output.set_property("foreground-rgba", color)

    def set_stderr_color(self, color):
        """Sets the color of the stderr text."""
        self._view.get_buffer().error.set_property("foreground-rgba", color)

    # pylint: disable=too-many-return-statements
    def __key_press_event_cb(self, view, event):
        buf = view.get_buffer()
        state = event.state & Gtk.accelerator_get_default_mod_mask()
        ctrl = state & Gdk.ModifierType.CONTROL_MASK

        if event.keyval == Gdk.KEY_Return:
            buf.process_command_line()
            return True

        if event.keyval in (Gdk.KEY_KP_Down, Gdk.KEY_Down):
            buf.history.down(buf.get_command_line())
            return True
        if event.keyval in (Gdk.KEY_KP_Up, Gdk.KEY_Up):
            buf.history.up(buf.get_command_line())
            return True

        if event.keyval in (Gdk.KEY_KP_Left, Gdk.KEY_Left, Gdk.KEY_BackSpace):
            return buf.is_cursor(at=True)

        if event.keyval in (Gdk.KEY_KP_Home, Gdk.KEY_Home):
            buf.place_cursor(buf.get_iter_at_mark(buf.prompt_mark))
            return True

        if (ctrl and event.keyval == Gdk.KEY_d) or event.keyval == Gdk.KEY_Escape:
            return self.emit("eof")

        return False

    def __mark_set_cb(self, buf, unused_iter, mark):
        if not mark.props.name == "insert":
            return

        self._view.set_editable(buf.is_cursor(at=True, after=True))

    def __insert_text_cb(self, buf, unused_iter, unused_text, unused_len):
        GLib.idle_add(self.scroll_to_end)
